@@ -142,8 +142,8 @@ def coarsen(reduction, x, axes, trim_excess=False):
         x = x[ind]
 
     # (10, 10) -> (5, 2, 5, 2)
-    newshape = tuple(concat([(x.shape[i] / axes[i], axes[i])
-                                for i in range(x.ndim)]))
+    newshape = tuple(concat([(x.shape[i] // axes[i], axes[i])
+                             for i in range(x.ndim)]))
 
     return reduction(x.reshape(newshape), axis=tuple(range(1, x.ndim * 2, 2)))
 
@@ -186,3 +186,8 @@ def topk(k, x):
     k = np.minimum(k, len(x))
     ind = np.argpartition(x, -k)[-k:]
     return np.sort(x[ind])[::-1]
+
+
+def arange(start, stop, step, length, dtype):
+    res = np.arange(start, stop, step, dtype)
+    return res[:-1] if len(res) > length else res
