@@ -1,10 +1,13 @@
 from __future__ import absolute_import, division, print_function
 
 from functools import partial
+import itertools
 
 import numpy as np
 
-from .core import Array, normalize_chunks
+from .core import Array, normalize_chunks, stack
+from .wrap import empty
+from . import chunk
 from ..base import tokenize
 
 
@@ -47,7 +50,8 @@ def linspace(start, stop, num=50, chunks=None, dtype=None):
 
     space = float(range_) / (num - 1)
 
-    dtype = dtype or np.linspace(0, 1, 1).dtype
+    if dtype is None:
+        dtype = np.linspace(0, 1, 1).dtype
 
     name = 'linspace-' + tokenize((start, stop, num, chunks, dtype))
 
@@ -118,11 +122,7 @@ def arange(*args, **kwargs):
     if dtype is None:
         dtype = np.arange(0, 1, step).dtype
 
-    range_ = stop - start
-    num = int(abs(range_ // step))
-    if (range_ % step) != 0:
-        num += 1
-
+    num = max(np.ceil((stop - start) / step), 0)
     chunks = normalize_chunks(chunks, (num,))
 
     name = 'arange-' + tokenize((start, stop, step, chunks, num))
@@ -132,8 +132,67 @@ def arange(*args, **kwargs):
     for i, bs in enumerate(chunks[0]):
         blockstart = start + (elem_count * step)
         blockstop = start + ((elem_count + bs) * step)
-        task = (np.arange, blockstart, blockstop, step, dtype)
+        task = (chunk.arange, blockstart, blockstop, step, bs, dtype)
         dsk[(name, i)] = task
         elem_count += bs
 
     return Array(dsk, name, chunks, dtype=dtype)
+
+
+def indices(dimensions, dtype=int, chunks=None):
+    """
+    Implements NumPy's ``indices`` for Dask Arrays.
+
+    Generates a grid of indices covering the dimensions provided.
+
+    The final array has the shape ``(len(dimensions), *dimensions)``. The
+    chunks are used to specify the chunking for axis 1 up to
+    ``len(dimensions)``. The 0th axis always has chunks of length 1.
+
+    Parameters
+    ----------
+    dimensions : sequence of ints
+        The shape of the index grid.
+    dtype : dtype, optional
+        Type to use for the array. Default is ``int``.
+    chunks : sequence of ints
+        The number of samples on each block. Note that the last block will have
+        fewer samples if ``len(array) % chunks != 0``.
+
+    Returns
+    -------
+    grid : dask array
+    """
+    if chunks is None:
+        raise ValueError("Must supply a chunks= keyword argument")
+
+    dimensions = tuple(dimensions)
+    dtype = np.dtype(dtype)
+    chunks = tuple(chunks)
+
+    if len(dimensions) != len(chunks):
+        raise ValueError("Need one more chunk than dimensions.")
+
+    grid = []
+    if np.prod(dimensions):
+        for i in range(len(dimensions)):
+            s = len(dimensions) * [None]
+            s[i] = slice(None)
+            s = tuple(s)
+
+            r = arange(dimensions[i], dtype=dtype, chunks=chunks[i])
+            r = r[s]
+
+            for j in itertools.chain(range(i), range(i + 1, len(dimensions))):
+                r = r.repeat(dimensions[j], axis=j)
+
+            grid.append(r)
+
+    if grid:
+        grid = stack(grid)
+    else:
+        grid = empty(
+            (len(dimensions),) + dimensions, dtype=dtype, chunks=(1,) + chunks
+        )
+
+    return grid
