@@ -2,15 +2,17 @@ from __future__ import absolute_import, division, print_function
 
 from operator import getitem
 from itertools import product
+from numbers import Integral
 
-from toolz import merge, pipe, concat, partition, partial
+from toolz import merge, pipe, concat, partial
 from toolz.curried import map
 
+from . import chunk, wrap
+from .core import Array, map_blocks, concatenate, concatenate3, reshapelist
+from .. import sharedict
 from ..base import tokenize
 from ..core import flatten
 from ..utils import concrete
-from .core import Array, map_blocks, concatenate, concatenate3
-from . import chunk, wrap
 
 
 def fractional_slice(task, axes):
@@ -79,20 +81,7 @@ def expand_key(k, dims):
 
     seq = list(product([k[0]], *[inds(i, ind)
                                  for i, ind in enumerate(k[1:])]))
-    return reshape(shape, seq)
-
-
-def reshape(shape, seq):
-    """ Reshape iterator to nested shape
-
-    >>> reshape((2, 3), range(6))
-    [[0, 1, 2], [3, 4, 5]]
-    """
-    if len(shape) == 1:
-        return list(seq)
-    else:
-        n = int(len(seq) / shape[0])
-        return [reshape(shape[1:], part) for part in partition(n, seq)]
+    return reshapelist(shape, seq)
 
 
 def ghost_internal(x, axes):
@@ -138,8 +127,10 @@ def ghost_internal(x, axes):
                 mid.append(bd + axes.get(i, 0) * 2)
             chunks.append(left + mid + right)
 
-    return Array(merge(interior_slices, ghost_blocks, x.dask),
-                 name, chunks, dtype=x._dtype)
+    dsk = merge(interior_slices, ghost_blocks)
+    dsk = sharedict.merge(x.dask, (name, dsk))
+
+    return Array(dsk, name, chunks, dtype=x.dtype)
 
 
 def trim_internal(x, axes):
@@ -237,7 +228,7 @@ def constant(x, axis, depth, value):
     chunks[axis] = (depth,)
 
     c = wrap.full(tuple(map(sum, chunks)), value,
-                  chunks=tuple(chunks), dtype=x._dtype)
+                  chunks=tuple(chunks), dtype=x.dtype)
 
     return concatenate([c, x, c], axis=axis)
 
@@ -369,9 +360,8 @@ def add_dummy_padding(x, depth, boundary):
     array([..., 0, 1, 2, 3, 4, 5, ...])
     """
     for k, v in boundary.items():
-        if v == 'none':
-            d = depth[k]
-
+        d = depth[k]
+        if v == 'none' and d > 0:
             empty_shape = list(x.shape)
             empty_shape[k] = d
 
@@ -405,10 +395,11 @@ def map_overlap(x, func, depth, boundary=None, trim=True, **kwargs):
 
 
 def coerce_depth(ndim, depth):
-    if isinstance(depth, int):
+    if isinstance(depth, Integral):
         depth = (depth,) * ndim
     if isinstance(depth, tuple):
         depth = dict(zip(range(ndim), depth))
+
     return depth
 
 
