@@ -1,6 +1,5 @@
 from __future__ import absolute_import, division, print_function
 
-from itertools import count
 from functools import wraps
 from collections import Iterator
 
@@ -9,6 +8,7 @@ from toolz import merge, merge_sorted
 
 from .core import Array
 from ..base import tokenize
+from .. import sharedict
 
 
 @wraps(np.percentile)
@@ -30,8 +30,6 @@ def _percentile(a, q, interpolation='linear'):
     return np.percentile(a, q, interpolation=interpolation)
 
 
-names = ('percentile-%d' % i for i in count(1))
-
 def percentile(a, q, interpolation='linear'):
     """ Approximate percentile of 1-D array
 
@@ -44,13 +42,19 @@ def percentile(a, q, interpolation='linear'):
     token = tokenize(a, list(q), interpolation)
     name = 'percentile_chunk-' + token
     dsk = dict(((name, i), (_percentile, (key), q, interpolation))
-            for i, key in enumerate(a._keys()))
+               for i, key in enumerate(a._keys()))
 
     name2 = 'percentile-' + token
     dsk2 = {(name2, 0): (merge_percentiles, q, [q] * len(a.chunks[0]),
                          sorted(dsk), a.chunks[0], interpolation)}
 
-    return Array(merge(a.dask, dsk, dsk2), name2, chunks=((len(q),),))
+    dtype = a.dtype
+    if np.issubdtype(dtype, np.integer):
+        dtype = (np.array([], dtype=dtype) / 0.5).dtype
+
+    dsk = merge(dsk, dsk2)
+    dsk = sharedict.merge(a.dask, (name2, dsk))
+    return Array(dsk, name2, chunks=((len(q),),), dtype=dtype)
 
 
 def merge_percentiles(finalq, qs, vals, Ns, interpolation='lower'):
@@ -150,7 +154,7 @@ def merge_percentiles(finalq, qs, vals, Ns, interpolation='lower'):
         elif interpolation == 'higher':
             rv = combined_vals[upper]
         elif interpolation == 'midpoint':
-            rv = 0.5*(combined_vals[lower] + combined_vals[upper])
+            rv = 0.5 * (combined_vals[lower] + combined_vals[upper])
         elif interpolation == 'nearest':
             lower_residual = np.abs(combined_q[lower] - desired_q)
             upper_residual = np.abs(combined_q[upper] - desired_q)
