@@ -7,8 +7,9 @@ import numpy as np
 
 from .core import (normalize_chunks, Array, slices_from_chunks,
                    broadcast_shapes, broadcast_to)
+from .. import sharedict
 from ..base import tokenize
-from ..utils import different_seeds, ignoring
+from ..utils import ignoring, random_state_data
 
 
 def doc_wraps(func):
@@ -36,7 +37,7 @@ class RandomState(object):
     >>> state = da.random.RandomState(1234)  # a seed
     >>> x = state.normal(10, 0.1, size=3, chunks=(2,))
     >>> x.compute()
-    array([  9.95487579,  10.02999135,  10.08498441])
+    array([ 10.01867852,  10.04812289,   9.89649746])
 
     See Also:
         np.random.RandomState
@@ -81,13 +82,14 @@ class RandomState(object):
         # Broadcast all arguments, get tiny versions as well
         # Start adding the relevant bits to the graph
         dsk = {}
+        dsks = []
         lookup = {}
         small_args = []
         for i, ar in enumerate(args):
             if isinstance(ar, (np.ndarray, Array)):
                 res = _broadcast_any(ar, size, chunks)
                 if isinstance(res, Array):
-                    dsk.update(res.dask)
+                    dsks.append(res.dask)
                     lookup[i] = res.name
                 elif isinstance(res, np.ndarray):
                     name = 'array-{}'.format(tokenize(res))
@@ -102,7 +104,7 @@ class RandomState(object):
             if isinstance(ar, (np.ndarray, Array)):
                 res = _broadcast_any(ar, size, chunks)
                 if isinstance(res, Array):
-                    dsk.update(res.dask)
+                    dsks.append(res.dask)
                     lookup[key] = res.name
                 elif isinstance(res, np.ndarray):
                     name = 'array-{}'.format(tokenize(res))
@@ -118,15 +120,15 @@ class RandomState(object):
                      **small_kwargs).dtype
 
         sizes = list(product(*chunks))
-        seeds = different_seeds(len(sizes), self._numpy_state)
-        token = tokenize(seeds, size, chunks, args, kwargs)
+        state_data = random_state_data(len(sizes), self._numpy_state)
+        token = tokenize(state_data, size, chunks, args, kwargs)
         name = 'da.random.{0}-{1}'.format(func.__name__, token)
 
         keys = product([name], *([range(len(bd)) for bd in chunks] +
                                  [[0]] * len(extra_chunks)))
         blocks = product(*[range(len(bd)) for bd in chunks])
         vals = []
-        for seed, size, slc, block in zip(seeds, sizes, slices, blocks):
+        for state, size, slc, block in zip(state_data, sizes, slices, blocks):
             arg = []
             for i, ar in enumerate(args):
                 if i not in lookup:
@@ -145,8 +147,9 @@ class RandomState(object):
                         kwrg[k] = (lookup[k], ) + block
                     else:   # np.ndarray
                         kwrg[k] = (getitem, lookup[k], slc)
-            vals.append((_apply_random, func.__name__, seed, size, arg, kwrg))
+            vals.append((_apply_random, func.__name__, state, size, arg, kwrg))
         dsk.update(dict(zip(keys, vals)))
+        dsk = sharedict.merge((name, dsk), *dsks)
         return Array(dsk, name, chunks + extra_chunks, dtype=dtype)
 
     @doc_wraps(np.random.RandomState.beta)
@@ -184,9 +187,9 @@ class RandomState(object):
                           size=size, chunks=chunks)
 
     @doc_wraps(np.random.RandomState.gamma)
-    def gamma(self, shape, scale=1.0, chunks=None):
-        return self._wrap(np.random.RandomState.gamma, scale,
-                          size=shape, chunks=chunks)
+    def gamma(self, shape, scale=1.0, size=None, chunks=None):
+        return self._wrap(np.random.RandomState.gamma, shape, scale,
+                          size=size, chunks=chunks)
 
     @doc_wraps(np.random.RandomState.geometric)
     def geometric(self, p, size=None, chunks=None):
@@ -349,13 +352,9 @@ class RandomState(object):
                           size=size, chunks=chunks)
 
 
-def _apply_random(func, seed, size, args, kwargs):
-    """ Apply RandomState method with seed
-
-    >>> _apply_random('normal', 123, 3, (10, 1.0), {})
-    array([  8.9143694 ,  10.99734545,  10.2829785 ])
-    """
-    state = np.random.RandomState(seed)
+def _apply_random(func, state_data, size, args, kwargs):
+    """Apply RandomState method with seed"""
+    state = np.random.RandomState(state_data)
     func = getattr(state, func)
     return func(*args, size=size, **kwargs)
 
