@@ -6,19 +6,22 @@ from math import factorial, log, ceil
 import operator
 
 import numpy as np
-from toolz import compose, partition_all, merge, get, accumulate, pluck
+from toolz import compose, partition_all, get, accumulate, pluck
 
 from . import chunk
-from .core import _concatenate2, Array, atop, sqrt, lol_tuples
+from .core import _concatenate2, Array, atop, lol_tuples, handle_out
+from .ufunc import sqrt
+from .wrap import zeros, ones
 from .numpy_compat import divide
 from ..compatibility import getargspec, builtins
 from ..base import tokenize
 from ..context import _globals
 from ..utils import ignoring, funcname
+from .. import sharedict
 
 
 def reduction(x, chunk, aggregate, axis=None, keepdims=None, dtype=None,
-              split_every=None, combine=None, name=None):
+              split_every=None, combine=None, name=None, out=None):
     """ General version of reductions
 
     >>> reduction(my_array, np.sum, np.sum, axis=0, keepdims=False)  # doctest: +SKIP
@@ -27,25 +30,29 @@ def reduction(x, chunk, aggregate, axis=None, keepdims=None, dtype=None,
         axis = tuple(range(x.ndim))
     if isinstance(axis, int):
         axis = (axis,)
-    axis = tuple(i if i >= 0 else x.ndim + i for i in axis)
+    axis = tuple(validate_axis(x.ndim, a) for a in axis)
 
-    if dtype and 'dtype' in getargspec(chunk).args:
+    if dtype is None:
+        raise ValueError("Must specify dtype")
+    if 'dtype' in getargspec(chunk).args:
         chunk = partial(chunk, dtype=dtype)
-    if dtype and 'dtype' in getargspec(aggregate).args:
+    if 'dtype' in getargspec(aggregate).args:
         aggregate = partial(aggregate, dtype=dtype)
 
     # Map chunk across all blocks
     inds = tuple(range(x.ndim))
-    tmp = atop(partial(chunk, axis=axis, keepdims=True), inds, x, inds)
-    tmp._chunks = tuple((1,)*len(c) if i in axis else c for (i, c)
+    # The dtype of `tmp` doesn't actually matter, and may be incorrect.
+    tmp = atop(chunk, inds, x, inds, axis=axis, keepdims=True, dtype=x.dtype)
+    tmp._chunks = tuple((1, ) * len(c) if i in axis else c for (i, c)
                         in enumerate(tmp.chunks))
 
-    return _tree_reduce(tmp, aggregate, axis, keepdims, dtype, split_every,
-                       combine, name=name)
+    result = _tree_reduce(tmp, aggregate, axis, keepdims, dtype, split_every,
+                          combine, name=name)
+    return handle_out(out, result)
 
 
 def _tree_reduce(x, aggregate, axis, keepdims, dtype, split_every=None,
-                combine=None, name=None):
+                 combine=None, name=None):
     """Perform the tree reduction step of a reduction.
 
     Lower level, users should use ``reduction`` or ``arg_reduction`` directly.
@@ -55,7 +62,7 @@ def _tree_reduce(x, aggregate, axis, keepdims, dtype, split_every=None,
     if isinstance(split_every, dict):
         split_every = dict((k, split_every.get(k, 2)) for k in axis)
     elif isinstance(split_every, int):
-        n = builtins.max(int(split_every ** (1/(len(axis) or 1))), 2)
+        n = builtins.max(int(split_every ** (1 / (len(axis) or 1))), 2)
         split_every = dict.fromkeys(axis, n)
     else:
         split_every = dict((k, v) for (k, v) in enumerate(x.numblocks) if k in axis)
@@ -68,12 +75,11 @@ def _tree_reduce(x, aggregate, axis, keepdims, dtype, split_every=None,
     func = compose(partial(combine or aggregate, axis=axis, keepdims=True),
                    partial(_concatenate2, axes=axis))
     for i in range(depth - 1):
-        x = partial_reduce(func, x, split_every, True, None,
+        x = partial_reduce(func, x, split_every, True, dtype=dtype,
                            name=(name or funcname(combine or aggregate)) + '-partial')
     func = compose(partial(aggregate, axis=axis, keepdims=keepdims),
                    partial(_concatenate2, axes=axis))
-    return partial_reduce(func, x, split_every, keepdims=keepdims,
-                          dtype=dtype,
+    return partial_reduce(func, x, split_every, keepdims=keepdims, dtype=dtype,
                           name=(name or funcname(aggregate)) + '-aggregate')
 
 
@@ -95,7 +101,7 @@ def partial_reduce(func, x, split_every, keepdims=False, dtype=None, name=None):
     >>> partial_reduce(np.min, x, {0: 1, 2: 3})    # doctest: +SKIP
     """
     name = (name or funcname(func)) + '-' + tokenize(func, x, split_every,
-                                                   keepdims, dtype)
+                                                     keepdims, dtype)
     parts = [list(partition_all(split_every.get(i, 1), range(n))) for (i, n)
              in enumerate(x.numblocks)]
     keys = product(*map(range, map(len, parts)))
@@ -112,102 +118,98 @@ def partial_reduce(func, x, split_every, keepdims=False, dtype=None, name=None):
         dummy = dict(i for i in enumerate(p) if i[0] not in decided)
         g = lol_tuples((x.name,), range(x.ndim), decided, dummy)
         dsk[(name,) + k] = (func, g)
-    return Array(merge(dsk, x.dask), name, out_chunks, dtype=dtype)
+    return Array(sharedict.merge(x.dask, (name, dsk)), name, out_chunks, dtype=dtype)
 
 
 @wraps(chunk.sum)
-def sum(a, axis=None, dtype=None, keepdims=False, split_every=None):
+def sum(a, axis=None, dtype=None, keepdims=False, split_every=None, out=None):
     if dtype is not None:
         dt = dtype
-    elif a._dtype is not None:
-        dt = np.empty((1,), dtype=a._dtype).sum().dtype
     else:
-        dt = None
+        dt = np.empty((1,), dtype=a.dtype).sum().dtype
     return reduction(a, chunk.sum, chunk.sum, axis=axis, keepdims=keepdims,
-                     dtype=dt, split_every=split_every)
+                     dtype=dt, split_every=split_every, out=out)
 
 
 @wraps(chunk.prod)
-def prod(a, axis=None, dtype=None, keepdims=False, split_every=None):
+def prod(a, axis=None, dtype=None, keepdims=False, split_every=None, out=None):
     if dtype is not None:
         dt = dtype
-    elif a._dtype is not None:
-        dt = np.empty((1,), dtype=a._dtype).prod().dtype
     else:
-        dt = None
+        dt = np.empty((1,), dtype=a.dtype).prod().dtype
     return reduction(a, chunk.prod, chunk.prod, axis=axis, keepdims=keepdims,
-                     dtype=dt, split_every=split_every)
+                     dtype=dt, split_every=split_every, out=out)
 
 
 @wraps(chunk.min)
-def min(a, axis=None, keepdims=False, split_every=None):
+def min(a, axis=None, keepdims=False, split_every=None, out=None):
     return reduction(a, chunk.min, chunk.min, axis=axis, keepdims=keepdims,
-                     dtype=a._dtype, split_every=split_every)
+                     dtype=a.dtype, split_every=split_every, out=out)
 
 
 @wraps(chunk.max)
-def max(a, axis=None, keepdims=False, split_every=None):
+def max(a, axis=None, keepdims=False, split_every=None, out=None):
     return reduction(a, chunk.max, chunk.max, axis=axis, keepdims=keepdims,
-                     dtype=a._dtype, split_every=split_every)
+                     dtype=a.dtype, split_every=split_every, out=out)
 
 
 @wraps(chunk.any)
-def any(a, axis=None, keepdims=False, split_every=None):
+def any(a, axis=None, keepdims=False, split_every=None, out=None):
     return reduction(a, chunk.any, chunk.any, axis=axis, keepdims=keepdims,
-                     dtype='bool', split_every=split_every)
+                     dtype='bool', split_every=split_every, out=out)
 
 
 @wraps(chunk.all)
-def all(a, axis=None, keepdims=False, split_every=None):
+def all(a, axis=None, keepdims=False, split_every=None, out=None):
     return reduction(a, chunk.all, chunk.all, axis=axis, keepdims=keepdims,
-                     dtype='bool', split_every=split_every)
+                     dtype='bool', split_every=split_every, out=out)
 
 
 @wraps(chunk.nansum)
-def nansum(a, axis=None, dtype=None, keepdims=False, split_every=None):
+def nansum(a, axis=None, dtype=None, keepdims=False, split_every=None, out=None):
     if dtype is not None:
         dt = dtype
-    elif a._dtype is not None:
-        dt = chunk.nansum(np.empty((1,), dtype=a._dtype)).dtype
     else:
-        dt = None
+        dt = chunk.nansum(np.empty((1,), dtype=a.dtype)).dtype
     return reduction(a, chunk.nansum, chunk.sum, axis=axis, keepdims=keepdims,
-                     dtype=dt, split_every=split_every)
+                     dtype=dt, split_every=split_every, out=out)
 
 
 with ignoring(AttributeError):
     @wraps(chunk.nanprod)
-    def nanprod(a, axis=None, dtype=None, keepdims=False, split_every=None):
+    def nanprod(a, axis=None, dtype=None, keepdims=False, split_every=None,
+                out=None):
         if dtype is not None:
             dt = dtype
-        elif a._dtype is not None:
-            dt = chunk.nanprod(np.empty((1,), dtype=a._dtype)).dtype
         else:
-            dt = None
+            dt = chunk.nanprod(np.empty((1,), dtype=a.dtype)).dtype
         return reduction(a, chunk.nanprod, chunk.prod, axis=axis,
-                         keepdims=keepdims, dtype=dt, split_every=split_every)
-
+                         keepdims=keepdims, dtype=dt, split_every=split_every,
+                         out=out)
 
     @wraps(chunk.nancumsum)
-    def nancumsum(x, axis, dtype=None):
-        return cumreduction(chunk.nancumsum, operator.add, 0, x, axis, dtype)
-
+    def nancumsum(x, axis, dtype=None, out=None):
+        return cumreduction(chunk.nancumsum, operator.add, 0, x, axis, dtype,
+                            out=out)
 
     @wraps(chunk.nancumprod)
-    def nancumprod(x, axis, dtype=None):
-        return cumreduction(chunk.nancumprod, operator.mul, 1, x, axis, dtype)
+    def nancumprod(x, axis, dtype=None, out=None):
+        return cumreduction(chunk.nancumprod, operator.mul, 1, x, axis, dtype,
+                            out=out)
 
 
 @wraps(chunk.nanmin)
-def nanmin(a, axis=None, keepdims=False, split_every=None):
+def nanmin(a, axis=None, keepdims=False, split_every=None, out=None):
     return reduction(a, chunk.nanmin, chunk.nanmin, axis=axis,
-                     keepdims=keepdims, dtype=a._dtype, split_every=split_every)
+                     keepdims=keepdims, dtype=a.dtype, split_every=split_every,
+                     out=out)
 
 
 @wraps(chunk.nanmax)
-def nanmax(a, axis=None, keepdims=False, split_every=None):
+def nanmax(a, axis=None, keepdims=False, split_every=None, out=None):
     return reduction(a, chunk.nanmax, chunk.nanmax, axis=axis,
-                     keepdims=keepdims, dtype=a._dtype, split_every=split_every)
+                     keepdims=keepdims, dtype=a.dtype, split_every=split_every,
+                     out=out)
 
 
 def numel(x, **kwargs):
@@ -224,7 +226,7 @@ def mean_chunk(x, sum=chunk.sum, numel=numel, dtype='f8', **kwargs):
     n = numel(x, dtype=dtype, **kwargs)
     total = sum(x, dtype=dtype, **kwargs)
     result = np.empty(shape=n.shape,
-              dtype=[('total', total.dtype), ('n', n.dtype)])
+                      dtype=[('total', total.dtype), ('n', n.dtype)])
     result['n'] = n
     result['total'] = total
     return result
@@ -245,43 +247,42 @@ def mean_agg(pair, dtype='f8', **kwargs):
 
 
 @wraps(chunk.mean)
-def mean(a, axis=None, dtype=None, keepdims=False, split_every=None):
+def mean(a, axis=None, dtype=None, keepdims=False, split_every=None, out=None):
     if dtype is not None:
         dt = dtype
-    elif a._dtype is not None:
-        dt = np.mean(np.empty(shape=(1,), dtype=a._dtype)).dtype
     else:
-        dt = None
+        dt = np.mean(np.empty(shape=(1,), dtype=a.dtype)).dtype
     return reduction(a, mean_chunk, mean_agg, axis=axis, keepdims=keepdims,
-                     dtype=dt, split_every=split_every, combine=mean_combine)
+                     dtype=dt, split_every=split_every, combine=mean_combine,
+                     out=out)
 
 
-def nanmean(a, axis=None, dtype=None, keepdims=False, split_every=None):
+def nanmean(a, axis=None, dtype=None, keepdims=False, split_every=None,
+            out=None):
     if dtype is not None:
         dt = dtype
-    elif a._dtype is not None:
-        dt = np.mean(np.empty(shape=(1,), dtype=a._dtype)).dtype
     else:
-        dt = None
+        dt = np.mean(np.empty(shape=(1,), dtype=a.dtype)).dtype
     return reduction(a, partial(mean_chunk, sum=chunk.nansum, numel=nannumel),
                      mean_agg, axis=axis, keepdims=keepdims, dtype=dt,
-                     split_every=split_every,
+                     split_every=split_every, out=out,
                      combine=partial(mean_combine, sum=chunk.nansum, numel=nannumel))
 
+
 with ignoring(AttributeError):
     nanmean = wraps(chunk.nanmean)(nanmean)
 
 
 def moment_chunk(A, order=2, sum=chunk.sum, numel=numel, dtype='f8', **kwargs):
     total = sum(A, dtype=dtype, **kwargs)
-    n = numel(A, **kwargs)
-    u = total/n
+    n = numel(A, **kwargs).astype(np.int64, copy=False)
+    u = total / n
     M = np.empty(shape=n.shape + (order - 1,), dtype=dtype)
     for i in range(2, order + 1):
         M[..., i - 2] = sum((A - u)**i, dtype=dtype, **kwargs)
     result = np.empty(shape=n.shape, dtype=[('total', total.dtype),
                                             ('n', n.dtype),
-                                            ('M', M.dtype, (order-1,))])
+                                            ('M', M.dtype, (order - 1,))])
     result['total'] = total
     result['n'] = n
     result['M'] = M
@@ -289,9 +290,9 @@ def moment_chunk(A, order=2, sum=chunk.sum, numel=numel, dtype='f8', **kwargs):
 
 
 def _moment_helper(Ms, ns, inner_term, order, sum, kwargs):
-    M = Ms[..., order - 2].sum(**kwargs) + sum(ns * inner_term**order, **kwargs)
+    M = Ms[..., order - 2].sum(**kwargs) + sum(ns * inner_term ** order, **kwargs)
     for k in range(1, order - 1):
-        coeff = factorial(order)/(factorial(k)*factorial(order - k))
+        coeff = factorial(order) / (factorial(k) * factorial(order - k))
         M += coeff * sum(Ms[..., order - k - 2] * inner_term**k, **kwargs)
     return M
 
@@ -314,7 +315,7 @@ def moment_combine(data, order=2, ddof=0, dtype='f8', sum=np.sum, **kwargs):
 
     result = np.zeros(shape=n.shape, dtype=[('total', total.dtype),
                                             ('n', n.dtype),
-                                            ('M', Ms.dtype, (order-1,))])
+                                            ('M', Ms.dtype, (order - 1,))])
     result['total'] = total
     result['n'] = n
     result['M'] = M
@@ -341,70 +342,82 @@ def moment_agg(data, order=2, ddof=0, dtype='f8', sum=np.sum, **kwargs):
 
 
 def moment(a, order, axis=None, dtype=None, keepdims=False, ddof=0,
-           split_every=None):
-    if not isinstance(order, int) or order < 2:
-        raise ValueError("Order must be an integer >= 2")
+           split_every=None, out=None):
+    if not isinstance(order, int) or order < 0:
+        raise ValueError("Order must be an integer >= 0")
+
+    if order < 2:
+        reduced = a.sum(axis=axis)   # get reduced shape and chunks
+        if order == 0:
+            # When order equals 0, the result is 1, by definition.
+            return ones(reduced.shape, chunks=reduced.chunks, dtype='f8')
+        # By definition the first order about the mean is 0.
+        return zeros(reduced.shape, chunks=reduced.chunks, dtype='f8')
+
     if dtype is not None:
         dt = dtype
-    elif a._dtype is not None:
-        dt = np.var(np.ones(shape=(1,), dtype=a._dtype)).dtype
     else:
-        dt = None
-    return reduction(a, partial(moment_chunk, order=order), partial(moment_agg,
-                     order=order, ddof=ddof), axis=axis, keepdims=keepdims,
-                     dtype=dt, split_every=split_every,
+        dt = np.var(np.ones(shape=(1,), dtype=a.dtype)).dtype
+    return reduction(a, partial(moment_chunk, order=order),
+                     partial(moment_agg, order=order, ddof=ddof),
+                     axis=axis, keepdims=keepdims,
+                     dtype=dt, split_every=split_every, out=out,
                      combine=partial(moment_combine, order=order))
 
 
 @wraps(chunk.var)
-def var(a, axis=None, dtype=None, keepdims=False, ddof=0, split_every=None):
+def var(a, axis=None, dtype=None, keepdims=False, ddof=0, split_every=None,
+        out=None):
     if dtype is not None:
         dt = dtype
-    elif a._dtype is not None:
-        dt = np.var(np.ones(shape=(1,), dtype=a._dtype)).dtype
     else:
-        dt = None
+        dt = np.var(np.ones(shape=(1,), dtype=a.dtype)).dtype
     return reduction(a, moment_chunk, partial(moment_agg, ddof=ddof), axis=axis,
                      keepdims=keepdims, dtype=dt, split_every=split_every,
-                     combine=moment_combine, name='var')
+                     combine=moment_combine, name='var', out=out)
 
 
-def nanvar(a, axis=None, dtype=None, keepdims=False, ddof=0, split_every=None):
+def nanvar(a, axis=None, dtype=None, keepdims=False, ddof=0, split_every=None,
+           out=None):
     if dtype is not None:
         dt = dtype
-    elif a._dtype is not None:
-        dt = np.var(np.ones(shape=(1,), dtype=a._dtype)).dtype
     else:
-        dt = None
+        dt = np.var(np.ones(shape=(1,), dtype=a.dtype)).dtype
     return reduction(a, partial(moment_chunk, sum=chunk.nansum, numel=nannumel),
                      partial(moment_agg, sum=np.nansum, ddof=ddof), axis=axis,
                      keepdims=keepdims, dtype=dt, split_every=split_every,
-                     combine=partial(moment_combine, sum=np.nansum))
+                     combine=partial(moment_combine, sum=np.nansum), out=out)
+
 
 with ignoring(AttributeError):
     nanvar = wraps(chunk.nanvar)(nanvar)
 
+
 @wraps(chunk.std)
-def std(a, axis=None, dtype=None, keepdims=False, ddof=0, split_every=None):
+def std(a, axis=None, dtype=None, keepdims=False, ddof=0, split_every=None,
+        out=None):
     result = sqrt(a.var(axis=axis, dtype=dtype, keepdims=keepdims, ddof=ddof,
-                        split_every=split_every))
+                        split_every=split_every, out=out))
     if dtype and dtype != result.dtype:
         result = result.astype(dtype)
     return result
 
 
-def nanstd(a, axis=None, dtype=None, keepdims=False, ddof=0, split_every=None):
+def nanstd(a, axis=None, dtype=None, keepdims=False, ddof=0, split_every=None,
+           out=None):
     result = sqrt(nanvar(a, axis=axis, dtype=dtype, keepdims=keepdims,
-                         ddof=ddof, split_every=split_every))
+                         ddof=ddof, split_every=split_every, out=out))
     if dtype and dtype != result.dtype:
         result = result.astype(dtype)
     return result
 
+
 with ignoring(AttributeError):
     nanstd = wraps(chunk.nanstd)(nanstd)
 
 
-def vnorm(a, ord=None, axis=None, dtype=None, keepdims=False, split_every=None):
+def vnorm(a, ord=None, axis=None, dtype=None, keepdims=False, split_every=None,
+          out=None):
     """ Vector norm
 
     See np.linalg.norm
@@ -412,18 +425,20 @@ def vnorm(a, ord=None, axis=None, dtype=None, keepdims=False, split_every=None):
     if ord is None or ord == 'fro':
         ord = 2
     if ord == np.inf:
-        return max(abs(a), axis=axis, keepdims=keepdims, split_every=split_every)
+        return max(abs(a), axis=axis, keepdims=keepdims,
+                   split_every=split_every, out=out)
     elif ord == -np.inf:
-        return min(abs(a), axis=axis, keepdims=keepdims, split_every=split_every)
+        return min(abs(a), axis=axis, keepdims=keepdims,
+                   split_every=split_every, out=out)
     elif ord == 1:
         return sum(abs(a), axis=axis, dtype=dtype, keepdims=keepdims,
-                   split_every=split_every)
+                   split_every=split_every, out=out)
     elif ord % 2 == 0:
-        return sum(a**ord, axis=axis, dtype=dtype, keepdims=keepdims,
-                   split_every=split_every)**(1./ord)
+        return sum(a ** ord, axis=axis, dtype=dtype, keepdims=keepdims,
+                   split_every=split_every, out=out) ** (1. / ord)
     else:
-        return sum(abs(a)**ord, axis=axis, dtype=dtype, keepdims=keepdims,
-                   split_every=split_every)**(1./ord)
+        return sum(abs(a) ** ord, axis=axis, dtype=dtype, keepdims=keepdims,
+                   split_every=split_every, out=out) ** (1. / ord)
 
 
 def _arg_combine(data, axis, argfunc, keepdims=False):
@@ -486,7 +501,7 @@ def nanarg_agg(func, argfunc, data, axis=None, **kwargs):
     return arg
 
 
-def arg_reduction(x, chunk, combine, agg, axis=None, split_every=None):
+def arg_reduction(x, chunk, combine, agg, axis=None, split_every=None, out=None):
     """Generic function for argreduction.
 
     Parameters
@@ -526,12 +541,15 @@ def arg_reduction(x, chunk, combine, agg, axis=None, split_every=None):
     else:
         offset_info = pluck(axis[0], offsets)
 
-    chunks = tuple((1,)*len(c) if i in axis else c for (i, c)
+    chunks = tuple((1, ) * len(c) if i in axis else c for (i, c)
                    in enumerate(x.chunks))
     dsk = dict(((name,) + k, (chunk, (old,) + k, axis, off)) for (k, off)
                in zip(keys, offset_info))
-    tmp = Array(merge(dsk, x.dask), name, chunks)
-    return _tree_reduce(tmp, agg, axis, False, np.int64, split_every, combine)
+    # The dtype of `tmp` doesn't actually matter, just need to provide something
+    tmp = Array(sharedict.merge(x.dask, (name, dsk)), name, chunks, dtype=x.dtype)
+    dtype = np.argmin([1]).dtype
+    result = _tree_reduce(tmp, agg, axis, False, dtype, split_every, combine)
+    return handle_out(out, result)
 
 
 def make_arg_reduction(func, argfunc, is_nan_func=False):
@@ -550,9 +568,12 @@ def make_arg_reduction(func, argfunc, is_nan_func=False):
         agg = partial(nanarg_agg, func, argfunc)
     else:
         agg = partial(arg_agg, func, argfunc)
+
     @wraps(argfunc)
-    def _(x, axis=None, split_every=None):
-        return arg_reduction(x, chunk, combine, agg, axis, split_every)
+    def _(x, axis=None, split_every=None, out=None):
+        return arg_reduction(x, chunk, combine, agg, axis,
+                             split_every=split_every, out=out)
+
     return _
 
 
@@ -576,7 +597,7 @@ nanargmin = make_arg_reduction(chunk.nanmin, _nanargmin, True)
 nanargmax = make_arg_reduction(chunk.nanmax, _nanargmax, True)
 
 
-def cumreduction(func, binop, ident, x, axis, dtype=None):
+def cumreduction(func, binop, ident, x, axis, dtype=None, out=None):
     """ Generic function for cumulative reduction
 
     Parameters
@@ -603,6 +624,8 @@ def cumreduction(func, binop, ident, x, axis, dtype=None):
     if dtype is None:
         dtype = func(np.empty((0,), dtype=x.dtype)).dtype
     assert isinstance(axis, int)
+    axis = validate_axis(x.ndim, axis)
+
     m = x.map_blocks(func, axis=axis, dtype=dtype)
 
     name = '%s-axis=%d-%s' % (func.__name__, axis, tokenize(x, dtype))
@@ -629,14 +652,26 @@ def cumreduction(func, binop, ident, x, axis, dtype=None):
                                       (operator.getitem, (m.name,) + old, slc))
             dsk[(name,) + ind] = (binop, this_slice, (m.name,) + ind)
 
-    return Array(merge(dsk, m.dask), name, x.chunks, m.dtype)
+    result = Array(sharedict.merge(m.dask, (name, dsk)), name, x.chunks, m.dtype)
+    return handle_out(out, result)
 
 
 @wraps(np.cumsum)
-def cumsum(x, axis, dtype=None):
-    return cumreduction(np.cumsum, operator.add, 0, x, axis, dtype)
+def cumsum(x, axis, dtype=None, out=None):
+    return cumreduction(np.cumsum, operator.add, 0, x, axis, dtype, out=out)
 
 
 @wraps(np.cumprod)
-def cumprod(x, axis, dtype=None):
-    return cumreduction(np.cumprod, operator.mul, 1, x, axis, dtype)
+def cumprod(x, axis, dtype=None, out=None):
+    return cumreduction(np.cumprod, operator.mul, 1, x, axis, dtype, out=out)
+
+
+def validate_axis(ndim, axis):
+    """ Validate single axis dimension against number of dimensions """
+    if axis > ndim - 1 or axis < -ndim:
+        raise ValueError("Axis must be between -%d and %d, got %d" %
+                         (ndim, ndim - 1, axis))
+    if axis < 0:
+        return axis + ndim
+    else:
+        return axis
