@@ -60,18 +60,18 @@ def test_arange():
     assert (sorted(da.arange(77, chunks=13, dtype=float).dask) ==
             sorted(da.arange(77, chunks=13, dtype=float).dask))
 
+    # 0 size output
+    darr = da.arange(0, 1, -0.5, chunks=20)
+    nparr = np.arange(0, 1, -0.5)
+    assert_eq(darr, nparr)
+
+    darr = da.arange(0, -1, 0.5, chunks=20)
+    nparr = np.arange(0, -1, 0.5)
+    assert_eq(darr, nparr)
+
 
 def test_arange_has_dtype():
-    assert da.arange(5, chunks=2)._dtype is not None
-
-
-def test_arange_working_float_step():
-    """Sometimes floating point step arguments work, but this could be platform
-    dependent.
-    """
-    darr = da.arange(3.3, -9.1, -.25, chunks=3)
-    nparr = np.arange(3.3, -9.1, -.25)
-    assert_eq(darr, nparr)
+    assert da.arange(5, chunks=2).dtype == np.arange(5).dtype
 
 
 @pytest.mark.xfail(reason="Casting floats to ints is not supported since edge"
@@ -82,8 +82,6 @@ def test_arange_cast_float_int_step():
     assert_eq(darr, nparr)
 
 
-@pytest.mark.xfail(reason="arange with a floating point step value can fail"
-                          "due to numerical instability.")
 def test_arange_float_step():
     darr = da.arange(2., 13., .3, chunks=4)
     nparr = np.arange(2., 13., .3)
@@ -92,3 +90,65 @@ def test_arange_float_step():
     darr = da.arange(7.7, 1.5, -.8, chunks=3)
     nparr = np.arange(7.7, 1.5, -.8)
     assert_eq(darr, nparr)
+
+    darr = da.arange(0, 1, 0.01, chunks=20)
+    nparr = np.arange(0, 1, 0.01)
+    assert_eq(darr, nparr)
+
+    darr = da.arange(0, 1, 0.03, chunks=20)
+    nparr = np.arange(0, 1, 0.03)
+    assert_eq(darr, nparr)
+
+
+def test_indices_no_chunks():
+    with pytest.raises(ValueError):
+        da.indices((1,))
+
+
+def test_indices_wrong_chunks():
+    with pytest.raises(ValueError):
+        da.indices((1,), chunks=tuple())
+
+
+def test_empty_indicies():
+    darr = da.indices(tuple(), chunks=tuple())
+    nparr = np.indices(tuple())
+    assert darr.shape == nparr.shape
+    assert darr.dtype == nparr.dtype
+    assert_eq(darr, nparr)
+
+    darr = da.indices(tuple(), float, chunks=tuple())
+    nparr = np.indices(tuple(), float)
+    assert darr.shape == nparr.shape
+    assert darr.dtype == nparr.dtype
+    assert_eq(darr, nparr)
+
+    darr = da.indices((0,), float, chunks=(1,))
+    nparr = np.indices((0,), float)
+    assert darr.shape == nparr.shape
+    assert darr.dtype == nparr.dtype
+    assert_eq(darr, nparr)
+
+    darr = da.indices((0, 1, 2), float, chunks=(1, 1, 2))
+    nparr = np.indices((0, 1, 2), float)
+    assert darr.shape == nparr.shape
+    assert darr.dtype == nparr.dtype
+    assert_eq(darr, nparr)
+
+
+def test_indicies():
+    darr = da.indices((1,), chunks=(1,))
+    nparr = np.indices((1,))
+    assert_eq(darr, nparr)
+
+    darr = da.indices((1,), float, chunks=(1,))
+    nparr = np.indices((1,), float)
+    assert_eq(darr, nparr)
+
+    darr = da.indices((2, 1), chunks=(2, 1))
+    nparr = np.indices((2, 1))
+    assert_eq(darr, nparr)
+
+    darr = da.indices((2, 3), chunks=(1, 2))
+    nparr = np.indices((2, 3))
+    assert_eq(darr, nparr)
