@@ -190,20 +190,22 @@ def test_lu_1():
         _check_lu_result(dp, dl, du, A)
 
 
+@pytest.mark.slow
 @pytest.mark.parametrize('size', [10, 20, 30, 50])
 def test_lu_2(size):
     np.random.seed(10)
-    A = np.random.random_integers(0, 10, (size, size))
+    A = np.random.randint(0, 10, (size, size))
 
     dA = da.from_array(A, chunks=(5, 5))
     dp, dl, du = da.linalg.lu(dA)
     _check_lu_result(dp, dl, du, A)
 
 
+@pytest.mark.slow
 @pytest.mark.parametrize('size', [50, 100, 200])
 def test_lu_3(size):
     np.random.seed(10)
-    A = np.random.random_integers(0, 10, (size, size))
+    A = np.random.randint(0, 10, (size, size))
 
     dA = da.from_array(A, chunks=(25, 25))
     dp, dl, du = da.linalg.lu(dA)
@@ -211,15 +213,15 @@ def test_lu_3(size):
 
 
 def test_lu_errors():
-    A = np.random.random_integers(0, 10, (10, 10, 10))
+    A = np.random.randint(0, 11, (10, 10, 10))
     dA = da.from_array(A, chunks=(5, 5, 5))
     pytest.raises(ValueError, lambda: da.linalg.lu(dA))
 
-    A = np.random.random_integers(0, 10, (10, 8))
+    A = np.random.randint(0, 11, (10, 8))
     dA = da.from_array(A, chunks=(5, 4))
     pytest.raises(ValueError, lambda: da.linalg.lu(dA))
 
-    A = np.random.random_integers(0, 10, (20, 20))
+    A = np.random.randint(0, 11, (20, 20))
     dA = da.from_array(A, chunks=(5, 4))
     pytest.raises(ValueError, lambda: da.linalg.lu(dA))
 
@@ -228,8 +230,8 @@ def test_lu_errors():
 def test_solve_triangular_vector(shape, chunk):
     np.random.seed(1)
 
-    A = np.random.random_integers(1, 10, (shape, shape))
-    b = np.random.random_integers(1, 10, shape)
+    A = np.random.randint(1, 11, (shape, shape))
+    b = np.random.randint(1, 11, shape)
 
     # upper
     Au = np.triu(A)
@@ -252,8 +254,8 @@ def test_solve_triangular_vector(shape, chunk):
 def test_solve_triangular_matrix(shape, chunk):
     np.random.seed(1)
 
-    A = np.random.random_integers(1, 10, (shape, shape))
-    b = np.random.random_integers(1, 10, (shape, 5))
+    A = np.random.randint(1, 10, (shape, shape))
+    b = np.random.randint(1, 10, (shape, 5))
 
     # upper
     Au = np.triu(A)
@@ -276,8 +278,8 @@ def test_solve_triangular_matrix(shape, chunk):
 def test_solve_triangular_matrix2(shape, chunk):
     np.random.seed(1)
 
-    A = np.random.random_integers(1, 10, (shape, shape))
-    b = np.random.random_integers(1, 10, (shape, shape))
+    A = np.random.randint(1, 10, (shape, shape))
+    b = np.random.randint(1, 10, (shape, shape))
 
     # upper
     Au = np.triu(A)
@@ -297,14 +299,14 @@ def test_solve_triangular_matrix2(shape, chunk):
 
 
 def test_solve_triangular_errors():
-    A = np.random.random_integers(0, 10, (10, 10, 10))
-    b = np.random.random_integers(1, 10, 10)
+    A = np.random.randint(0, 10, (10, 10, 10))
+    b = np.random.randint(1, 10, 10)
     dA = da.from_array(A, chunks=(5, 5, 5))
     db = da.from_array(b, chunks=5)
     pytest.raises(ValueError, lambda: da.linalg.solve_triangular(dA, db))
 
-    A = np.random.random_integers(0, 10, (10, 10))
-    b = np.random.random_integers(1, 10, 10)
+    A = np.random.randint(0, 10, (10, 10))
+    b = np.random.randint(1, 10, 10)
     dA = da.from_array(A, chunks=(3, 3))
     db = da.from_array(b, chunks=5)
     pytest.raises(ValueError, lambda: da.linalg.solve_triangular(dA, db))
@@ -314,11 +316,11 @@ def test_solve_triangular_errors():
 def test_solve(shape, chunk):
     np.random.seed(1)
 
-    A = np.random.random_integers(1, 10, (shape, shape))
+    A = np.random.randint(1, 10, (shape, shape))
     dA = da.from_array(A, (chunk, chunk))
 
     # vector
-    b = np.random.random_integers(1, 10, shape)
+    b = np.random.randint(1, 10, shape)
     db = da.from_array(b, chunk)
 
     res = da.linalg.solve(dA, db)
@@ -326,7 +328,7 @@ def test_solve(shape, chunk):
     assert_eq(dA.dot(res), b.astype(float))
 
     # tall-and-skinny matrix
-    b = np.random.random_integers(1, 10, (shape, 5))
+    b = np.random.randint(1, 10, (shape, 5))
     db = da.from_array(b, (chunk, 5))
 
     res = da.linalg.solve(dA, db)
@@ -334,7 +336,7 @@ def test_solve(shape, chunk):
     assert_eq(dA.dot(res), b.astype(float))
 
     # matrix
-    b = np.random.random_integers(1, 10, (shape, shape))
+    b = np.random.randint(1, 10, (shape, shape))
     db = da.from_array(b, (chunk, chunk))
 
     res = da.linalg.solve(dA, db)
@@ -346,7 +348,7 @@ def test_solve(shape, chunk):
 def test_inv(shape, chunk):
     np.random.seed(1)
 
-    A = np.random.random_integers(1, 10, (shape, shape))
+    A = np.random.randint(1, 10, (shape, shape))
     dA = da.from_array(A, (chunk, chunk))
 
     res = da.linalg.inv(dA)
@@ -356,7 +358,7 @@ def test_inv(shape, chunk):
 
 def _get_symmat(size):
     np.random.seed(1)
-    A = np.random.random_integers(1, 20, (size, size))
+    A = np.random.randint(1, 21, (size, size))
     lA = np.tril(A)
     return lA.dot(lA.T)
 
@@ -369,7 +371,7 @@ def test_solve_sym_pos(shape, chunk):
     dA = da.from_array(A, (chunk, chunk))
 
     # vector
-    b = np.random.random_integers(1, 10, shape)
+    b = np.random.randint(1, 10, shape)
     db = da.from_array(b, chunk)
 
     res = da.linalg.solve(dA, db, sym_pos=True)
@@ -377,7 +379,7 @@ def test_solve_sym_pos(shape, chunk):
     assert_eq(dA.dot(res), b.astype(float))
 
     # tall-and-skinny matrix
-    b = np.random.random_integers(1, 10, (shape, 5))
+    b = np.random.randint(1, 10, (shape, 5))
     db = da.from_array(b, (chunk, 5))
 
     res = da.linalg.solve(dA, db, sym_pos=True)
@@ -385,7 +387,7 @@ def test_solve_sym_pos(shape, chunk):
     assert_eq(dA.dot(res), b.astype(float))
 
     # matrix
-    b = np.random.random_integers(1, 10, (shape, shape))
+    b = np.random.randint(1, 10, (shape, shape))
     db = da.from_array(b, (chunk, chunk))
 
     res = da.linalg.solve(dA, db, sym_pos=True)
@@ -406,8 +408,8 @@ def test_cholesky(shape, chunk):
                          [(20, 10, 5), (100, 10, 10)])
 def test_lstsq(nrow, ncol, chunk):
     np.random.seed(1)
-    A = np.random.random_integers(1, 20, (nrow, ncol))
-    b = np.random.random_integers(1, 20, nrow)
+    A = np.random.randint(1, 20, (nrow, ncol))
+    b = np.random.randint(1, 20, nrow)
 
     dA = da.from_array(A, (chunk, ncol))
     db = da.from_array(b, chunk)
@@ -424,7 +426,32 @@ def test_lstsq(nrow, ncol, chunk):
     A[:, 1] = A[:, 2]
     dA = da.from_array(A, (chunk, ncol))
     db = da.from_array(b, chunk)
-    x, r, rank, s = np.linalg.lstsq(A, b)
+    x, r, rank, s = np.linalg.lstsq(A, b,
+                                    rcond=np.finfo(np.double).eps * max(nrow,
+                                                                        ncol))
     assert rank == ncol - 1
     dx, dr, drank, ds = da.linalg.lstsq(dA, db)
     assert drank.compute() == rank
+
+
+def test_no_chunks_svd():
+    x = np.random.random((100, 10))
+    u, s, v = np.linalg.svd(x, full_matrices=0)
+
+    for chunks in [((np.nan,) * 10, (10,)),
+                   ((np.nan,) * 10, (np.nan,))]:
+        dx = da.from_array(x, chunks=(10, 10))
+        dx._chunks = chunks
+
+        du, ds, dv = da.linalg.svd(dx)
+
+        assert_eq(s, ds)
+        assert_eq(u.dot(np.diag(s)).dot(v),
+                  du.dot(da.diag(ds)).dot(dv))
+        assert_eq(du.T.dot(du), np.eye(10))
+        assert_eq(dv.T.dot(dv), np.eye(10))
+
+        dx = da.from_array(x, chunks=(10, 10))
+        dx._chunks = ((np.nan,) * 10, (np.nan,))
+        assert_eq(abs(v), abs(dv))
+        assert_eq(abs(u), abs(du))
