@@ -4,9 +4,10 @@ from __future__ import absolute_import, division, print_function
 import pytest
 import math
 import os
+import random
 import sys
 from collections import Iterator
-from distutils.version import LooseVersion
+from itertools import repeat
 
 import partd
 from toolz import merge, join, filter, identity, valmap, groupby, pluck
@@ -14,11 +15,10 @@ from toolz import merge, join, filter, identity, valmap, groupby, pluck
 import dask
 import dask.bag as db
 from dask.bag.core import (Bag, lazify, lazify_task, map, collect,
-        reduceby, reify, partition, inline_singleton_lists, optimize,
-        from_delayed)
-from dask.async import get_sync
+                           reduceby, reify, partition, inline_singleton_lists,
+                           optimize, from_delayed)
 from dask.compatibility import BZ2File, GzipFile, PY2
-from dask.utils import filetexts, tmpfile, tmpdir, raises, open
+from dask.utils import filetexts, tmpfile, tmpdir
 from dask.utils_test import inc, add
 
 
@@ -48,14 +48,14 @@ def test_keys():
     assert sorted(b._keys()) == sorted(dsk.keys())
 
 
+# Deprecated test for old map behavior
 def test_map():
     c = b.map(inc)
-    expected = merge(dsk, dict(((c.name, i), (reify, (map, inc, (b.name, i))))
-                               for i in range(b.npartitions)))
-    assert c.dask == expected
+    assert c.compute() == list(map(inc, b.compute()))
     assert c.name == b.map(inc).name
 
 
+# Deprecated test for old map behavior
 def test_map_function_with_multiple_arguments():
     b = db.from_sequence([(1, 10), (2, 20), (3, 30)], npartitions=3)
     assert list(b.map(lambda x, y: x + y).compute(get=dask.get)) == [11, 22, 33]
@@ -72,6 +72,7 @@ class B(object):
         pass
 
 
+# Deprecated test for old map behavior
 def test_map_with_constructors():
     assert db.from_sequence([[1, 2, 3]]).map(A).compute()
     assert db.from_sequence([1, 2, 3]).map(B).compute()
@@ -85,6 +86,7 @@ def test_map_with_constructors():
     assert failed
 
 
+# Deprecated test for old map behavior
 def test_map_with_builtins():
     b = db.from_sequence(range(3))
     assert ' '.join(b.map(str)) == '0 1 2'
@@ -99,6 +101,7 @@ def test_map_with_builtins():
         [True, False]
 
 
+# Deprecated test for old map behavior
 def test_map_with_kwargs():
     b = db.from_sequence(range(100), npartitions=10)
     assert b.map(lambda x, factor=0: x * factor,
@@ -110,6 +113,105 @@ def test_map_with_kwargs():
                  factor=2).sum().compute() == 2.0
 
 
+def test_bag_map():
+    b = db.from_sequence(range(100), npartitions=10)
+    b2 = db.from_sequence(range(100, 200), npartitions=10)
+    x = b.compute()
+    x2 = b2.compute()
+
+    def myadd(a=1, b=2, c=3):
+        return a + b + c
+
+    assert db.map(myadd, b).compute() == list(map(myadd, x))
+    assert db.map(myadd, a=b).compute() == list(map(myadd, x))
+    assert db.map(myadd, b, b2).compute() == list(map(myadd, x, x2))
+    assert db.map(myadd, b, 10).compute() == [myadd(i, 10) for i in x]
+    assert db.map(myadd, 10, b=b).compute() == [myadd(10, b=i) for i in x]
+
+    sol = [myadd(i, b=j, c=100) for (i, j) in zip(x, x2)]
+    assert db.map(myadd, b, b=b2, c=100).compute() == sol
+
+    sol = [myadd(i, c=100) for (i, j) in zip(x, x2)]
+    assert db.map(myadd, b, c=100).compute() == sol
+
+    x_sum = sum(x)
+    sol = [myadd(x_sum, b=i, c=100) for i in x2]
+    assert db.map(myadd, b.sum(), b=b2, c=100).compute() == sol
+
+    sol = [myadd(i, b=x_sum, c=100) for i in x2]
+    assert db.map(myadd, b2, b.sum(), c=100).compute() == sol
+
+    sol = [myadd(a=100, b=x_sum, c=i) for i in x2]
+    assert db.map(myadd, a=100, b=b.sum(), c=b2).compute() == sol
+
+    a = dask.delayed(10)
+    assert db.map(myadd, b, a).compute() == [myadd(i, 10) for i in x]
+    assert db.map(myadd, b, b=a).compute() == [myadd(i, b=10) for i in x]
+
+    # Mispatched npartitions
+    fewer_parts = db.from_sequence(range(100), npartitions=5)
+    with pytest.raises(ValueError):
+        db.map(myadd, b, fewer_parts)
+
+    # No bags
+    with pytest.raises(ValueError):
+        db.map(myadd, b.sum(), 1, 2)
+
+    # Unequal partitioning
+    unequal = db.from_sequence(range(110), npartitions=10)
+    with pytest.raises(ValueError):
+        db.map(myadd, b, unequal, c=b2).compute()
+    with pytest.raises(ValueError):
+        db.map(myadd, b, b=unequal, c=b2).compute()
+
+
+def test_map_method():
+    b = db.from_sequence(range(100), npartitions=10)
+    b2 = db.from_sequence(range(100, 200), npartitions=10)
+    x = b.compute()
+    x2 = b2.compute()
+
+    def myadd(a, b=2, c=3):
+        return a + b + c
+
+    assert b.map(myadd).compute() == list(map(myadd, x))
+    assert b.map(myadd, b2).compute() == list(map(myadd, x, x2))
+    assert b.map(myadd, 10).compute() == [myadd(i, 10) for i in x]
+    assert b.map(myadd, b=10).compute() == [myadd(i, b=10) for i in x]
+    assert (b.map(myadd, b2, c=10).compute() ==
+            [myadd(i, j, 10) for (i, j) in zip(x, x2)])
+    x_sum = sum(x)
+    assert (b.map(myadd, b.sum(), c=10).compute() ==
+            [myadd(i, x_sum, 10) for i in x])
+
+    # check that map works with multiarg functions. Can be removed after
+    # deprecated behavior is removed
+    assert b.map(add, b2).compute() == list(map(add, x, x2))
+
+    # check that map works with vararg functions. Can be removed after
+    # deprecated behavior is removed
+    def vararg_inc(*args):
+        return inc(*args)
+
+    assert b.map(vararg_inc).compute(get=dask.get) == list(map(inc, x))
+
+
+def test_starmap():
+    data = [(1, 2), (3, 4), (5, 6), (7, 8), (9, 10)]
+    b = db.from_sequence(data, npartitions=2)
+
+    def myadd(a, b, c=0):
+        return a + b + c
+
+    assert b.starmap(myadd).compute() == [myadd(*a) for a in data]
+    assert b.starmap(myadd, c=10).compute() == [myadd(*a, c=10) for a in data]
+    max_second = b.pluck(1).max()
+    assert (b.starmap(myadd, c=max_second).compute() ==
+            [myadd(*a, c=max_second.compute()) for a in data])
+    c = dask.delayed(10)
+    assert b.starmap(myadd, c=c).compute() == [myadd(*a, c=10) for a in data]
+
+
 def test_filter():
     c = b.filter(iseven)
     expected = merge(dsk, dict(((c.name, i),
@@ -149,7 +251,7 @@ def test_pluck():
 
 def test_pluck_with_default():
     b = db.from_sequence(['Hello', '', 'World'])
-    assert raises(IndexError, lambda: list(b.pluck(0)))
+    pytest.raises(IndexError, lambda: list(b.pluck(0)))
     assert list(b.pluck(0, None)) == ['H', None, 'W']
     assert b.pluck(0, None).name == b.pluck(0, None).name
     assert b.pluck(0).name != b.pluck(0, None).name
@@ -189,7 +291,7 @@ def test_fold():
     assert set(d.fold(lambda a, b: ''.join([a, b]), initial='').compute()) == set('hello')
 
     e = db.from_sequence([[1], [2], [3]], npartitions=2)
-    with dask.set_options(get=get_sync):
+    with dask.set_options(get=dask.get):
         assert set(e.fold(add, initial=[]).compute()) == set([1, 2, 3])
 
 
@@ -198,6 +300,8 @@ def test_distinct():
     assert b.distinct().name == b.distinct().name
     assert 'distinct' in b.distinct().name
     assert b.distinct().count().compute() == 5
+    bag = db.from_sequence([0] * 50, npartitions=50)
+    assert bag.filter(None).distinct().compute() == []
 
 
 def test_frequencies():
@@ -208,6 +312,14 @@ def test_frequencies():
     assert c.name == b.frequencies().name
     assert c.name != c2.name
     assert c2.name == b.frequencies(split_every=2).name
+    # test bag with empty partitions
+    b2 = db.from_sequence(range(20), partition_size=2)
+    b2 = b2.filter(lambda x: x < 10)
+    d = b2.frequencies()
+    assert dict(d) == dict(zip(range(10), [1] * 10))
+    bag = db.from_sequence([0, 0, 0, 0], npartitions=4)
+    bag2 = bag.filter(None).frequencies(split_every=2)
+    assert dict(bag2.compute(get=dask.get)) == {}
 
 
 def test_topk():
@@ -220,8 +332,9 @@ def test_topk():
     assert b.topk(4).name == b.topk(4).name
 
 
-def test_topk_with_non_callable_key():
-    b = db.from_sequence([(1, 10), (2, 9), (3, 8)], npartitions=2)
+@pytest.mark.parametrize('npartitions', [1, 2])
+def test_topk_with_non_callable_key(npartitions):
+    b = db.from_sequence([(1, 10), (2, 9), (3, 8)], npartitions=npartitions)
     assert list(b.topk(2, key=1)) == [(1, 10), (2, 9)]
     assert list(b.topk(2, key=0)) == [(3, 8), (2, 9)]
     assert b.topk(2, key=1).name == b.topk(2, key=1).name
@@ -276,15 +389,20 @@ def test_tree_reductions():
     assert c.key != b.sum().key
 
 
-def test_mean():
-    assert b.mean().compute(get=dask.get) == 2.0
-    assert float(b.mean()) == 2.0
+@pytest.mark.parametrize('npartitions', [1, 3, 4])
+def test_aggregation(npartitions):
+    L = list(range(15))
+    b = db.range(15, npartitions=npartitions)
+    assert b.mean().compute(get=dask.get) == sum(L) / len(L)
+    assert b.sum().compute(get=dask.get) == sum(L)
+    assert b.count().compute(get=dask.get) == len(L)
 
 
-def test_non_splittable_reductions():
+@pytest.mark.parametrize('npartitions', [1, 10])
+def test_non_splittable_reductions(npartitions):
     np = pytest.importorskip('numpy')
     data = list(range(100))
-    c = db.from_sequence(data, npartitions=10)
+    c = db.from_sequence(data, npartitions=npartitions)
     assert c.mean().compute() == np.mean(data)
     assert c.std().compute(get=dask.get) == np.std(data)
 
@@ -302,8 +420,8 @@ def test_var():
 def test_join():
     c = b.join([1, 2, 3], on_self=isodd, on_other=iseven)
     assert list(c) == list(join(iseven, [1, 2, 3], isodd, list(b)))
-    assert list(b.join([1, 2, 3], isodd)) == \
-            list(join(isodd, [1, 2, 3], isodd, list(b)))
+    assert (list(b.join([1, 2, 3], isodd)) ==
+            list(join(isodd, [1, 2, 3], isodd, list(b))))
     assert c.name == b.join([1, 2, 3], on_self=isodd, on_other=iseven).name
 
 
@@ -323,18 +441,35 @@ def test_map_partitions():
     assert b.map_partitions(lambda a: len(a) + 1).name != b.map_partitions(len).name
 
 
-def test_map_partitions_with_kwargs():
-    b = db.from_sequence(range(100), npartitions=10)
-    assert b.map_partitions(
-        lambda X, factor=0: [x * factor for x in X],
-        factor=2).sum().compute() == 9900.0
-    assert b.map_partitions(
-        lambda X, total=0: [x / total for x in X],
-        total=b.sum()).sum().compute() == 1.0
-    assert b.map_partitions(
-        lambda X, factor=0, total=0: [x * factor / total for x in X],
-        total=b.sum(),
-        factor=2).sum().compute() == 2.0
+def test_map_partitions_args_kwargs():
+    x = [random.randint(-100, 100) for i in range(100)]
+    y = [random.randint(-100, 100) for i in range(100)]
+
+    dx = db.from_sequence(x, npartitions=10)
+    dy = db.from_sequence(y, npartitions=10)
+
+    def maximum(x, y=0):
+        y = repeat(y) if isinstance(y, int) else y
+        return [max(a, b) for (a, b) in zip(x, y)]
+
+    sol = maximum(x, y=10)
+    assert db.map_partitions(maximum, dx, y=10).compute() == sol
+    assert dx.map_partitions(maximum, y=10).compute() == sol
+    assert dx.map_partitions(maximum, 10).compute() == sol
+
+    sol = maximum(x, y)
+    assert db.map_partitions(maximum, dx, dy).compute() == sol
+    assert dx.map_partitions(maximum, y=dy).compute() == sol
+    assert dx.map_partitions(maximum, dy).compute() == sol
+
+    dy_mean = dy.mean().apply(int)
+    sol = maximum(x, int(sum(y) / len(y)))
+    assert dx.map_partitions(maximum, y=dy_mean).compute() == sol
+    assert dx.map_partitions(maximum, dy_mean).compute() == sol
+
+    dy_mean = dask.delayed(dy_mean)
+    assert dx.map_partitions(maximum, y=dy_mean).compute() == sol
+    assert dx.map_partitions(maximum, dy_mean).compute() == sol
 
 
 def test_random_sample_size():
@@ -395,10 +530,8 @@ def test_lazify_task():
     task = (reify, (map, inc, [1, 2, 3]))
     assert lazify_task(task) == task
 
-    a = (reify, (map, inc,
-                      (reify, (filter, iseven, 'y'))))
-    b = (reify, (map, inc,
-                              (filter, iseven, 'y')))
+    a = (reify, (map, inc, (reify, (filter, iseven, 'y'))))
+    b = (reify, (map, inc, (filter, iseven, 'y')))
     assert lazify_task(a) == b
 
 
@@ -406,11 +539,9 @@ f = lambda x: x
 
 
 def test_lazify():
-    a = {'x': (reify, (map, inc,
-                            (reify, (filter, iseven, 'y')))),
+    a = {'x': (reify, (map, inc, (reify, (filter, iseven, 'y')))),
          'a': (f, 'x'), 'b': (f, 'x')}
-    b = {'x': (reify, (map, inc,
-                                    (filter, iseven, 'y'))),
+    b = {'x': (reify, (map, inc, (filter, iseven, 'y'))),
          'a': (f, 'x'), 'b': (f, 'x')}
     assert lazify(a) == b
 
@@ -422,7 +553,7 @@ def test_inline_singleton_lists():
     assert inline_singleton_lists(inp) == out
 
     out = {'c': (f, 'a', 1)}
-    assert optimize(inp, ['c']) == out
+    assert optimize(inp, ['c'], rename_fused_keys=False) == out
 
     inp = {'b': (list, 'a'),
            'c': (f, 'b', 1),
@@ -448,7 +579,7 @@ def test_take_npartitions():
 
 
 @pytest.mark.skipif(sys.version_info[:2] == (3,3),
-    reason="Python3.3 uses pytest2.7.2, w/o warns method")
+                    reason="Python3.3 uses pytest2.7.2, w/o warns method")
 def test_take_npartitions_warn():
     with pytest.warns(None):
         b.take(100)
@@ -469,32 +600,9 @@ def test_can_use_dict_to_make_concrete():
     assert isinstance(dict(b.frequencies()), dict)
 
 
-def test_from_castra():
-    pytest.importorskip('castra')
-    pd = pytest.importorskip('pandas')
-    dd = pytest.importorskip('dask.dataframe')
-    blosc = pytest.importorskip('blosc')
-    if LooseVersion(blosc.__version__) == '1.3.0':
-        pytest.skip()
-    df = pd.DataFrame({'x': list(range(100)),
-                       'y': [str(i) for i in range(100)]})
-    a = dd.from_pandas(df, 10)
-
-    with tmpfile('.castra') as fn:
-        c = a.to_castra(fn)
-        default = db.from_castra(c)
-        with_columns = db.from_castra(c, 'x')
-        with_index = db.from_castra(c, 'x', index=True)
-        assert (list(default) == [{'x': i, 'y': str(i)}
-                                 for i in range(100)] or
-                list(default) == [(i, str(i)) for i in range(100)])
-        assert list(with_columns) == list(range(100))
-        assert list(with_index) == list(zip(range(100), range(100)))
-        assert default.name != with_columns.name != with_index.name
-        assert with_index.name == db.from_castra(c, 'x', index=True).name
-
-
 @pytest.mark.slow
+@pytest.mark.network
+@pytest.mark.skip(reason="Hangs")
 def test_from_url():
     a = db.from_url(['http://google.com', 'http://github.com'])
     assert a.npartitions == 2
@@ -506,12 +614,12 @@ def test_from_url():
 
 def test_read_text():
     with filetexts({'a1.log': 'A\nB', 'a2.log': 'C\nD'}) as fns:
-        assert set(line.strip() for line in db.read_text(fns)) == \
-                set('ABCD')
-        assert set(line.strip() for line in db.read_text('a*.log')) == \
-                set('ABCD')
+        assert (set(line.strip() for line in db.read_text(fns)) ==
+                set('ABCD'))
+        assert (set(line.strip() for line in db.read_text('a*.log')) ==
+                set('ABCD'))
 
-    assert raises(ValueError, lambda: db.read_text('non-existent-*-path'))
+    pytest.raises(ValueError, lambda: db.read_text('non-existent-*-path'))
 
 
 def test_read_text_large():
@@ -521,7 +629,7 @@ def test_read_text_large():
         b = db.read_text(fn, blocksize=100)
         c = db.read_text(fn)
         assert len(b.dask) > 5
-        assert list(map(str, b)) == list(map(str, c))
+        assert list(map(str, b.str.strip())) == list(map(str, c.str.strip()))
 
         d = db.read_text([fn], blocksize=100)
         assert list(b) == list(d)
@@ -534,7 +642,8 @@ def test_read_text_encoding():
         b = db.read_text(fn, blocksize=100, encoding='gb18030')
         c = db.read_text(fn, encoding='gb18030')
         assert len(b.dask) > 5
-        assert list(map(lambda x: x.encode('utf-8'), b)) == list(map(lambda x: x.encode('utf-8'), c))
+        assert (list(b.str.strip().map(lambda x: x.encode('utf-8'))) ==
+                list(c.str.strip().map(lambda x: x.encode('utf-8'))))
 
         d = db.read_text([fn], blocksize=100, encoding='gb18030')
         assert list(b) == list(d)
@@ -547,13 +656,14 @@ def test_read_text_large_gzip():
         f.close()
 
         with pytest.raises(ValueError):
-            db.read_text(fn, blocksize=100, linedelimiter='\n')
+            db.read_text(fn, blocksize=50, linedelimiter='\n')
 
         c = db.read_text(fn)
         assert c.npartitions == 1
 
 
 @pytest.mark.slow
+@pytest.mark.network
 def test_from_s3():
     # note we don't test connection modes with aws_access_key and
     # aws_secret_key because these are not on travis-ci
@@ -607,8 +717,7 @@ def test_partition_collect():
         assert set(p.get(1)) == set([1, 4])
         assert set(p.get(2)) == set([2, 5])
 
-        assert sorted(collect(identity, 0, p, '')) == \
-                [(0, [0]), (3, [3])]
+        assert sorted(collect(identity, 0, p, '')) == [(0, [0]), (3, [3])]
 
 
 def test_groupby():
@@ -648,13 +757,14 @@ def test_concat():
     b = db.from_sequence([4, 5, 6])
     c = db.concat([a, b])
     assert list(c) == [1, 2, 3, 4, 5, 6]
-
     assert c.name == db.concat([a, b]).name
-    assert b.concat().name != a.concat().name
-    assert b.concat().name == b.concat().name
 
-    b = db.from_sequence([1, 2, 3]).map(lambda x: x * [1, 2, 3])
-    assert list(b.concat()) == [1, 2, 3] * sum([1, 2, 3])
+
+def test_flatten():
+    b = db.from_sequence([[1], [2, 3]])
+    assert list(b.flatten()) == [1, 2, 3]
+    assert list(b.flatten()) == list(b.concat())
+    assert b.flatten().name == b.flatten().name
 
 
 def test_concat_after_map():
@@ -673,58 +783,82 @@ def test_args():
 
 
 def test_to_dataframe():
-    pytest.importorskip('dask.dataframe')
+    dd = pytest.importorskip('dask.dataframe')
     pd = pytest.importorskip('pandas')
-    b = db.from_sequence([(1, 2), (10, 20), (100, 200)], npartitions=2)
 
+    def check_parts(df, sol):
+        assert all((p.dtypes == sol.dtypes).all() for p in
+                   dask.compute(*df.to_delayed()))
+
+    dsk = {('test', 0): [(1, 2)],
+           ('test', 1): [],
+           ('test', 2): [(10, 20), (100, 200)]}
+    b = Bag(dsk, 'test', 3)
+    sol = pd.DataFrame(b.compute(), columns=['a', 'b'])
+
+    # Elements are tuples
     df = b.to_dataframe()
-    assert list(df.columns) == list(pd.DataFrame(list(b)).columns)
-
+    dd.utils.assert_eq(df, sol.rename(columns={'a': 0, 'b': 1}),
+                       check_index=False)
     df = b.to_dataframe(columns=['a', 'b'])
-    assert df.npartitions == b.npartitions
-    assert list(df.columns) == ['a', 'b']
+    dd.utils.assert_eq(df, sol, check_index=False)
+    check_parts(df, sol)
+    df = b.to_dataframe(meta=[('a', 'i8'), ('b', 'i8')])
+    dd.utils.assert_eq(df, sol, check_index=False)
+    check_parts(df, sol)
 
-    assert df.a.compute().values.tolist() == list(b.pluck(0))
-    assert df.b.compute().values.tolist() == list(b.pluck(1))
+    # Elements are dictionaries
+    b = b.map(lambda x: dict(zip(['a', 'b'], x)))
+    df = b.to_dataframe()
+    dd.utils.assert_eq(df, sol, check_index=False)
+    check_parts(df, sol)
+    assert df._name == b.to_dataframe()._name
 
-    b = db.from_sequence([{'a':   1, 'b':   2},
-                          {'a':  10, 'b':  20},
-                          {'a': 100, 'b': 200}], npartitions=2)
+    # With metadata specified
+    for meta in [sol, [('a', 'i8'), ('b', 'i8')]]:
+        df = b.to_dataframe(meta=meta)
+        dd.utils.assert_eq(df, sol, check_index=False)
+        check_parts(df, sol)
 
-    df2 = b.to_dataframe()
+    # Error to specify both columns and meta
+    with pytest.raises(ValueError):
+        b.to_dataframe(columns=['a', 'b'], meta=sol)
 
-    assert (df2.compute().values == df.compute().values).all()
+    # Single column
+    b = b.pluck('a')
+    sol = sol[['a']]
+    df = b.to_dataframe(meta=sol)
+    dd.utils.assert_eq(df, sol, check_index=False)
+    check_parts(df, sol)
 
-    assert df2._name == b.to_dataframe()._name
-    assert df2._name != df._name
-
-    meta = pd.DataFrame({'a': [1], 'b': [2]}).iloc[0:0]
-    df3 = b.to_dataframe(columns=meta)
-    assert df2._name == df3._name
-    assert (df3.compute().values == df2.compute().values).all()
-
-    b = db.from_sequence([1, 2, 3, 4, 5], npartitions=2)
-    df4 = b.to_dataframe()
-    assert len(df4.columns) == 1
-    assert list(df4.compute()) == list(pd.DataFrame(list(b)))
+    # Works with iterators and tuples
+    sol = pd.DataFrame({'a': range(100)})
+    b = db.from_sequence(range(100), npartitions=5)
+    for f in [iter, tuple]:
+        df = b.map_partitions(f).to_dataframe(meta=sol)
+        dd.utils.assert_eq(df, sol, check_index=False)
+        check_parts(df, sol)
 
 
-def test_to_textfiles():
+ext_open = [('gz', GzipFile), ('', open)]
+if not PY2:
+    ext_open.append(('bz2', BZ2File))
+
+
+@pytest.mark.parametrize('ext,myopen', ext_open)
+def test_to_textfiles(ext, myopen):
     b = db.from_sequence(['abc', '123', 'xyz'], npartitions=2)
-    for ext, myopen in [('gz', GzipFile), ('bz2', BZ2File), ('', open)]:
-        if ext == 'bz2' and PY2:
-            continue
-        with tmpdir() as dir:
-            c = b.to_textfiles(os.path.join(dir, '*.' + ext), compute=False)
-            dask.compute(*c)
-            assert os.path.exists(os.path.join(dir, '1.' + ext))
+    with tmpdir() as dir:
+        c = b.to_textfiles(os.path.join(dir, '*.' + ext), compute=False)
+        dask.compute(*c, get=dask.get)
+        assert os.path.exists(os.path.join(dir, '1.' + ext))
 
-            f = myopen(os.path.join(dir, '1.' + ext), 'rb')
-            text = f.read()
-            if hasattr(text, 'decode'):
-                text = text.decode()
-            assert 'xyz' in text
-            f.close()
+        f = myopen(os.path.join(dir, '1.' + ext), 'rb')
+        text = f.read()
+        if hasattr(text, 'decode'):
+            text = text.decode()
+        assert 'xyz' in text
+        f.close()
 
 
 def test_to_textfiles_name_function_preserves_order():
@@ -776,7 +910,7 @@ def test_to_textfiles_inputs():
         B.to_textfiles(dirname)
         assert os.path.exists(dirname)
         assert os.path.exists(os.path.join(dirname, '0.part'))
-    assert raises(ValueError, lambda: B.to_textfiles(5))
+    pytest.raises(ValueError, lambda: B.to_textfiles(5))
 
 
 def test_to_textfiles_endlines():
@@ -801,7 +935,7 @@ def test_string_namespace():
                                       ['Charlie', 'Smith']]
     assert list(b.str.match('*Smith')) == ['Alice Smith', 'Charlie Smith']
 
-    assert raises(AttributeError, lambda: b.str.sfohsofhf)
+    pytest.raises(AttributeError, lambda: b.str.sfohsofhf)
     assert b.str.match('*Smith').name == b.str.match('*Smith').name
     assert b.str.match('*Smith').name != b.str.match('*John').name
 
@@ -885,6 +1019,23 @@ def test_to_delayed():
     assert t.compute() == 21
 
 
+def test_to_delayed_optimizes():
+    b = db.from_sequence([1, 2, 3, 4, 5, 6], npartitions=1)
+    b2 = b.map(inc).map(inc).map(inc)
+
+    [d] = b2.to_delayed()
+    text = str(dict(d.dask))
+    assert text.count('reify') == 1
+
+    d = b2.sum().to_delayed()
+    text = str(dict(d.dask))
+    assert text.count('reify') == 0
+
+    [d] = b2.to_textfiles('foo.txt', compute=False)
+    text = str(dict(d.dask))
+    assert text.count('reify') <= 0
+
+
 def test_from_delayed():
     from dask.delayed import delayed
     a, b, c = delayed([1, 2, 3]), delayed([4, 5, 6]), delayed([7, 8, 9])
@@ -899,6 +1050,22 @@ def test_from_delayed():
     assert asum_value.compute() == asum_item.compute() == 6
 
 
+def test_from_delayed_iterator():
+    from dask.delayed import delayed
+
+    def lazy_records(n):
+        return ({'operations': [1, 2]} for _ in range(n))
+
+    delayed_records = delayed(lazy_records, pure=False)
+    bag = db.from_delayed([delayed_records(5) for _ in range(5)])
+    assert db.compute(
+        bag.count(),
+        bag.pluck('operations').count(),
+        bag.pluck('operations').flatten().count(),
+        get=dask.get,
+    ) == (25, 25, 50)
+
+
 def test_range():
     for npartitions in [1, 7, 10, 28]:
         b = db.range(100, npartitions=npartitions)
@@ -916,19 +1083,29 @@ def test_zip(npartitions, hi=1000):
     assert list(pairs) == list(zip(range(0, hi, 2), range(1, hi, 2)))
 
 
-def test_repartition():
-    for x, y in [(10, 5), (7, 3), (5, 1), (5, 4)]:
-        b = db.from_sequence(range(20), npartitions=x)
-        c = b.repartition(y)
+@pytest.mark.parametrize('nin', [1, 2, 7, 11, 23])
+@pytest.mark.parametrize('nout', [1, 2, 5, 12, 23])
+def test_repartition(nin, nout):
+    b = db.from_sequence(range(100), npartitions=nin)
+    c = b.repartition(npartitions=nout)
 
-        assert b.npartitions == x
-        assert c.npartitions == y
-        assert list(b) == c.compute(get=dask.get)
+    assert c.npartitions == nout
+    assert b.compute(get=dask.get) == c.compute(get=dask.get)
+    results = dask.get(c.dask, c._keys())
+    assert all(results)
 
-    try:
-        b.repartition(100)
-    except NotImplementedError as e:
-        assert '100' in str(e)
+
+def test_repartition_names():
+    b = db.from_sequence(range(100), npartitions=5)
+    c = b.repartition(2)
+    assert b.name != c.name
+
+    d = b.repartition(20)
+    assert b.name != c.name
+    assert c.name != d.name
+
+    c = b.repartition(5)
+    assert b is c
 
 
 @pytest.mark.skipif('not db.core._implement_accumulate')
@@ -1019,6 +1196,17 @@ def test_reduction_empty():
     assert b.filter(lambda x: x % 2 == 0).min().compute(get=dask.get) == 0
 
 
+@pytest.mark.parametrize('npartitions', [1, 2, 4])
+def test_reduction_empty_aggregate(npartitions):
+    b = db.from_sequence([0, 0, 0, 1], npartitions=npartitions).filter(None)
+    assert b.min(split_every=2).compute(get=dask.get) == 1
+    vals = db.compute(b.min(split_every=2), b.max(split_every=2), get=dask.get)
+    assert vals == (1, 1)
+    with pytest.raises(ValueError):
+        b = db.from_sequence([0, 0, 0, 0], npartitions=npartitions)
+        b.filter(None).min(split_every=2).compute(get=dask.get)
+
+
 class StrictReal(int):
     def __eq__(self, other):
         assert isinstance(other, StrictReal)
@@ -1062,7 +1250,7 @@ def test_bag_picklable():
 
 def test_msgpack_unicode():
     b = db.from_sequence([{"a": 1}]).groupby("a")
-    result = b.compute(get=dask.async.get_sync)
+    result = b.compute(get=dask.get)
     assert dict(result) == {1: [{'a': 1}]}
 
 
@@ -1070,3 +1258,57 @@ def test_bag_with_single_callable():
     f = lambda: None
     b = db.from_sequence([f])
     assert list(b.compute(get=dask.get)) == [f]
+
+
+def test_optimize_fuse_keys():
+    x = db.range(10, npartitions=2)
+    y = x.map(inc)
+    z = y.map(inc)
+
+    dsk = z._optimize(z.dask, z._keys())
+    assert not set(y.dask) & set(dsk)
+
+    dsk = z._optimize(z.dask, z._keys(), fuse_keys=y._keys())
+    assert all(k in dsk for k in y._keys())
+
+
+def test_reductions_are_lazy():
+    current = [None]
+
+    def part():
+        for i in range(10):
+            current[0] = i
+            yield i
+
+    def func(part):
+        assert current[0] == 0
+        return sum(part)
+
+    b = Bag({('foo', 0): part()}, 'foo', 1)
+
+    res = b.reduction(func, sum)
+
+    assert res.compute(get=dask.get) == sum(range(10))
+
+
+def test_repeated_groupby():
+    b = db.range(10, npartitions=4)
+    c = b.groupby(lambda x: x % 3)
+    assert valmap(len, dict(c)) == valmap(len, dict(c))
+
+
+def test_temporary_directory():
+    b = db.range(10, npartitions=4)
+
+    with dask.set_options(temporary_directory=os.getcwd()):
+        b2 = b.groupby(lambda x: x % 2)
+        b2.compute()
+        assert any(fn.endswith('.partd') for fn in os.listdir(os.getcwd()))
+
+
+def test_empty_bag():
+    b = db.from_sequence([])
+    assert b.map(inc).all().compute(get=dask.get)
+    assert not b.map(inc).any().compute(get=dask.get)
+    assert not b.map(inc).sum().compute(get=dask.get)
+    assert not b.map(inc).count().compute(get=dask.get)
