@@ -1,28 +1,31 @@
 from __future__ import print_function, division, absolute_import
 
+import gzip
 import os
 from time import sleep
+import sys
 
 import pytest
 from toolz import concat, valmap, partial
 
 from dask import compute, get, delayed
-from dask.compatibility import FileNotFoundError
+from dask.compatibility import FileNotFoundError, unicode
 from dask.utils import filetexts
 from dask.bytes import compression
-from dask.bytes.local import read_bytes, open_files, getsize, open_file_write
-from dask.bytes.core import open_text_files, write_bytes
+from dask.bytes.local import LocalFileSystem
+from dask.bytes.core import (open_text_files, write_bytes, read_bytes,
+                             open_files, OpenFileCreator)
 
 compute = partial(compute, get=get)
 
-files = {'.test.accounts.1.json':  (b'{"amount": 100, "name": "Alice"}\n'
-                              b'{"amount": 200, "name": "Bob"}\n'
-                              b'{"amount": 300, "name": "Charlie"}\n'
-                              b'{"amount": 400, "name": "Dennis"}\n'),
-         '.test.accounts.2.json':  (b'{"amount": 500, "name": "Alice"}\n'
-                              b'{"amount": 600, "name": "Bob"}\n'
-                              b'{"amount": 700, "name": "Charlie"}\n'
-                              b'{"amount": 800, "name": "Dennis"}\n')}
+files = {'.test.accounts.1.json': (b'{"amount": 100, "name": "Alice"}\n'
+                                   b'{"amount": 200, "name": "Bob"}\n'
+                                   b'{"amount": 300, "name": "Charlie"}\n'
+                                   b'{"amount": 400, "name": "Dennis"}\n'),
+         '.test.accounts.2.json': (b'{"amount": 500, "name": "Alice"}\n'
+                                   b'{"amount": 600, "name": "Bob"}\n'
+                                   b'{"amount": 700, "name": "Charlie"}\n'
+                                   b'{"amount": 800, "name": "Dennis"}\n')}
 
 
 def test_read_bytes():
@@ -60,6 +63,44 @@ def test_read_bytes_blocksize_none():
         assert sum(map(len, values)) == len(files)
 
 
+def test_read_bytes_blocksize_float():
+    with filetexts(files, mode='b'):
+        sample, vals = read_bytes('.test.account*', blocksize=5.0)
+        results = compute(*concat(vals))
+        ourlines = b"".join(results).split(b'\n')
+        testlines = b"".join(files.values()).split(b'\n')
+        assert set(ourlines) == set(testlines)
+
+        with pytest.raises(TypeError):
+            read_bytes('.test.account*', blocksize=5.5)
+
+
+def test_with_urls():
+    with filetexts(files, mode='b'):
+        # OS-independent file:// URI with glob *
+        url = to_uri('.test.accounts.') + '*'
+        sample, values = read_bytes(url, blocksize=None)
+        assert sum(map(len, values)) == len(files)
+
+
+@pytest.mark.skipif(sys.platform == 'win32',
+                    reason="pathlib and moto clash on windows")
+def test_with_paths():
+    pathlib = pytest.importorskip('pathlib')
+    with filetexts(files, mode='b'):
+        url = pathlib.Path('./.test.accounts.*')
+        sample, values = read_bytes(url, blocksize=None)
+        assert sum(map(len, values)) == len(files)
+    with pytest.raises(OSError):
+        # relative path doesn't work
+        url = pathlib.Path('file://.test.accounts.*')
+        read_bytes(url, blocksize=None)
+    pytest.importorskip("s3fs")
+    with pytest.raises(ValueError):
+        url = pathlib.Path('s3://bucket/test.accounts.*')
+        sample, values = read_bytes(url, blocksize=None)
+
+
 def test_read_bytes_block():
     with filetexts(files, mode='b'):
         for bs in [5, 15, 45, 1500]:
@@ -80,7 +121,7 @@ def test_read_bytes_delimited():
     with filetexts(files, mode='b'):
         for bs in [5, 15, 45, 1500]:
             _, values = read_bytes('.test.accounts*',
-                                    blocksize=bs, delimiter=b'\n')
+                                   blocksize=bs, delimiter=b'\n')
             _, values2 = read_bytes('.test.accounts*',
                                     blocksize=bs, delimiter=b'foo')
             assert ([a.key for a in concat(values)] !=
@@ -115,7 +156,8 @@ def test_compression(fmt, blocksize):
     files2 = valmap(compress, files)
     with filetexts(files2, mode='b'):
         sample, values = read_bytes('.test.accounts.*.json',
-                blocksize=blocksize, delimiter=b'\n', compression=fmt)
+                                    blocksize=blocksize, delimiter=b'\n',
+                                    compression=fmt)
         assert sample[:5] == files[sorted(files)[0]][:5]
         assert sample.endswith(b'\n')
 
@@ -134,11 +176,13 @@ def test_registered_read_bytes():
 
 
 def test_registered_open_files():
-    from dask.bytes.core import open_files
     with filetexts(files, mode='b'):
         myfiles = open_files('.test.accounts.*')
         assert len(myfiles) == len(files)
-        data = compute(*[file.read() for file in myfiles])
+        data = []
+        for file in myfiles:
+            with file as f:
+                data.append(f.read())
         assert list(data) == [files[k] for k in sorted(files)]
 
 
@@ -148,7 +192,10 @@ def test_registered_open_text_files(encoding):
     with filetexts(files, mode='b'):
         myfiles = open_text_files('.test.accounts.*', encoding=encoding)
         assert len(myfiles) == len(files)
-        data = compute(*[file.read() for file in myfiles])
+        data = []
+        for file in myfiles:
+            with file as f:
+                data.append(f.read())
         assert list(data) == [files[k].decode(encoding)
                               for k in sorted(files)]
 
@@ -157,17 +204,21 @@ def test_open_files():
     with filetexts(files, mode='b'):
         myfiles = open_files('.test.accounts.*')
         assert len(myfiles) == len(files)
-        data = compute(*[file.read() for file in myfiles])
-        assert list(data) == [files[k] for k in sorted(files)]
+        for lazy_file, data_file in zip(myfiles, sorted(files)):
+            with lazy_file as f:
+                x = f.read()
+                assert x == files[data_file]
 
 
 @pytest.mark.parametrize('fmt', list(compression.files))
 def test_compression_binary(fmt):
-    from dask.bytes.core import open_files
     files2 = valmap(compression.compress[fmt], files)
     with filetexts(files2, mode='b'):
         myfiles = open_files('.test.accounts.*', compression=fmt)
-        data = compute(*[file.read() for file in myfiles])
+        data = []
+        for file in myfiles:
+            with file as f:
+                data.append(f.read())
         assert list(data) == [files[k] for k in sorted(files)]
 
 
@@ -176,7 +227,10 @@ def test_compression_text(fmt):
     files2 = valmap(compression.compress[fmt], files)
     with filetexts(files2, mode='b'):
         myfiles = open_text_files('.test.accounts.*', compression=fmt)
-        data = compute(*[file.read() for file in myfiles])
+        data = []
+        for file in myfiles:
+            with file as f:
+                data.append(f.read())
         assert list(data) == [files[k].decode() for k in sorted(files)]
 
 
@@ -184,11 +238,11 @@ def test_compression_text(fmt):
 def test_getsize(fmt):
     compress = compression.compress[fmt]
     with filetexts({'.tmp.getsize': compress(b'1234567890')}, mode='b'):
-        assert getsize('.tmp.getsize', fmt) == 10
+        fs = LocalFileSystem()
+        assert fs.logical_size('.tmp.getsize', fmt) == 10
 
 
 def test_bad_compression():
-    from dask.bytes.core import read_bytes, open_files, open_text_files
     with filetexts(files, mode='b'):
         for func in [read_bytes, open_files, open_text_files]:
             with pytest.raises(ValueError):
@@ -198,7 +252,7 @@ def test_bad_compression():
 
 def test_not_found():
     fn = 'not-a-file'
-    with pytest.raises(FileNotFoundError) as e:
+    with pytest.raises((FileNotFoundError, OSError)) as e:
         read_bytes(fn)
     assert fn in str(e)
 
@@ -223,23 +277,6 @@ def test_names():
         assert [aa._key for aa in a] != [cc._key for cc in c]
 
 
-@pytest.mark.parametrize('open_files', [open_files, open_text_files])
-def test_modification_time_open_files(open_files):
-    with filetexts(files, mode='b'):
-        a = open_files('.test.accounts.*')
-        b = open_files('.test.accounts.*')
-
-        assert [aa._key for aa in a] == [bb._key for bb in b]
-
-    sleep(1)
-
-    double = lambda x: x + x
-    with filetexts(valmap(double, files), mode='b'):
-        c = open_files('.test.accounts.*')
-
-    assert [aa._key for aa in a] != [cc._key for cc in c]
-
-
 def test_simple_write(tmpdir):
     tmpdir = str(tmpdir)
     make_bytes = lambda: b'000'
@@ -273,8 +310,72 @@ def test_compressed_write(tmpdir):
 
 def test_open_files_write(tmpdir):
     tmpdir = str(tmpdir)
-    f = open_file_write([os.path.join(tmpdir, 'test1'),
-                         os.path.join(tmpdir, 'test2')])
-    assert len(f) == 2
-    files = compute(*f)
+    files = open_files([os.path.join(tmpdir, 'test1'),
+                        os.path.join(tmpdir, 'test2')], mode='wb')
+    assert len(files) == 2
     assert files[0].mode == 'wb'
+
+
+def test_pickability_of_lazy_files(tmpdir):
+    cloudpickle = pytest.importorskip('cloudpickle')
+    fn = os.path.join(str(tmpdir), 'foo')
+    with open(fn, 'wb') as f:
+        f.write(b'1')
+
+    opener = OpenFileCreator('file://foo.py', open=open)
+    opener2 = cloudpickle.loads(cloudpickle.dumps(opener))
+    assert type(opener2.fs) == type(opener.fs)
+
+    lazy_file = opener(fn, mode='rt')
+    lazy_file2 = cloudpickle.loads(cloudpickle.dumps(lazy_file))
+    assert lazy_file.path == lazy_file2.path
+
+    with lazy_file as f:
+        pass
+
+    lazy_file3 = cloudpickle.loads(cloudpickle.dumps(lazy_file))
+    assert lazy_file.path == lazy_file3.path
+
+
+def test_py2_local_bytes(tmpdir):
+    fn = str(tmpdir / 'myfile.txt.gz')
+    with gzip.open(fn, mode='wb') as f:
+        f.write(b'hello\nworld')
+
+    ofc = OpenFileCreator(fn, text=True, open=open, mode='rt',
+                          compression='gzip', encoding='utf-8')
+    lazy_file = ofc(fn)
+
+    with lazy_file as f:
+        assert all(isinstance(line, unicode) for line in f)
+
+
+def test_abs_paths(tmpdir):
+    tmpdir = str(tmpdir)
+    here = os.getcwd()
+    os.chdir(tmpdir)
+    with open('tmp', 'w') as f:
+        f.write('hi')
+    out = LocalFileSystem().glob('*')
+    assert len(out) == 1
+    assert os.sep in out[0]
+    assert tmpdir in out[0] and 'tmp' in out[0]
+
+    fs = LocalFileSystem()
+    os.chdir(here)
+    assert fs.open('tmp', 'r').read() == 'hi'
+
+
+try:
+    # used only in test_with_urls - may be more generally useful
+    import pathlib
+
+    def to_uri(path):
+        return pathlib.Path(os.path.abspath(path)).as_uri()
+
+except (ImportError, NameError):
+    import urlparse, urllib
+
+    def to_uri(path):
+        return urlparse.urljoin(
+            'file:', urllib.pathname2url(os.path.abspath(path)))
