@@ -1,11 +1,11 @@
 from __future__ import print_function, division, absolute_import
 
 from .core import (DataFrame, Series, Index, _Frame, map_partitions,
-                   repartition)
+                   repartition, to_delayed, to_datetime)
 from .io import (from_array, from_pandas, from_bcolz,
-                 from_dask_array, from_castra, read_hdf,
+                 from_dask_array, read_hdf, read_sql_table,
                  from_delayed, read_csv, to_csv, read_table,
-                 demo)
+                 demo, to_hdf, to_records, to_bag)
 from .optimize import optimize
 from .multi import merge, concat
 from .rolling import (rolling_count, rolling_sum, rolling_mean, rolling_median,
@@ -14,3 +14,7 @@ from .rolling import (rolling_count, rolling_sum, rolling_mean, rolling_median,
                       rolling_window)
 from ..base import compute
 from .reshape import get_dummies, pivot_table, melt
+try:
+    from .io import read_parquet, to_parquet
+except ImportError:
+    pass
