@@ -8,25 +8,134 @@ import warnings
 import numpy as np
 import pandas as pd
 
-from .core import DataFrame, Series, aca, map_partitions, no_default
+from .core import (DataFrame, Series, aca, map_partitions, merge,
+                   new_dd_object, no_default, split_out_on_index)
+from .methods import drop_columns
 from .shuffle import shuffle
 from .utils import make_meta, insert_meta_param_description, raise_on_meta_error
 from ..base import tokenize
 from ..utils import derived_from, M, funcname
 
 
+# #############################################
+#
+# GroupBy implementation notes
+#
+# Dask groupby supports reductions, i.e., mean, sum and alike, and apply. The
+# former do not shuffle the data and are efficiently implemented as tree
+# reductions. The latter is implemented by shuffling the underlying partiitons
+# such that all items of a group can be found in the same parititon.
+#
+# The argument to ``.groupby``, the index, can be a ``str``, ``dd.DataFrame``,
+# ``dd.Series``, or a list thereof. In operations on the grouped object, the
+# divisions of the the grouped object and the items of index have to align.
+# Currently, there is no support to shuffle the index values as part of the
+# groupby operation. Therefore, the alignment has to be guaranteed by the
+# caller.
+#
+# To operate on matchings paritions, most groupby operations exploit the
+# corresponding support in ``apply_concat_apply``. Specifically, this function
+# operates on matching paritiotns of frame-like objects passed as varargs.
+#
+# After the inital chunk step, the passed index is implicitly passed along to
+# subsequent operations as the index of the parittions. Groupby operations on
+# the individual parttions can then access the index via the ``levels``
+# parameter of the ``groupby`` function. The correct arguments is determined by
+# the ``_determine_levels`` function.
+#
+# To minimize overhead, series in an index that were obtained by getitem on the
+# object to group are not passed as series to the various operations, but as
+# columnn keys. This transformation is implemented as ``_normalize_index``.
+#
+# #############################################
+
+def _determine_levels(index):
+    """Determine the correct levels argument to groupby.
+    """
+    if isinstance(index, (tuple, list)) and len(index) > 1:
+        return list(range(len(index)))
+    else:
+        return 0
+
+
+def _normalize_index(df, index):
+    """Replace series with column names in an index wherever possible.
+    """
+    if not isinstance(df, DataFrame):
+        return index
+
+    elif isinstance(index, list):
+        return [_normalize_index(df, col) for col in index]
+
+    elif (isinstance(index, Series) and index.name in df.columns and
+          index._name == df[index.name]._name):
+        return index.name
+
+    elif (isinstance(index, DataFrame) and
+          set(index.columns).issubset(df.columns) and
+          index._name == df[index.columns]._name):
+        return list(index.columns)
+
+    else:
+        return index
+
+
 def _maybe_slice(grouped, columns):
     """
     Slice columns if grouped is pd.DataFrameGroupBy
     """
     if isinstance(grouped, pd.core.groupby.DataFrameGroupBy):
         if columns is not None:
-            columns = columns if isinstance(columns, str) else list(columns)
+            if isinstance(columns, (tuple, list, set, pd.Index)):
+                columns = list(columns)
             return grouped[columns]
     return grouped
 
 
+def _is_aligned(df, by):
+    """Check if `df` and `by` have aligned indices"""
+    if isinstance(by, (pd.Series, pd.DataFrame)):
+        return df.index.equals(by.index)
+    elif isinstance(by, (list, tuple)):
+        return all(_is_aligned(df, i) for i in by)
+    else:
+        return True
+
+
+def _groupby_raise_unaligned(df, **kwargs):
+    """Groupby, but raise if df and `by` key are unaligned.
+
+    Pandas supports grouping by a column that doesn't align with the input
+    frame/series/index. However, the reindexing this causes doesn't seem to be
+    threadsafe, and can result in incorrect results. Since grouping by an
+    unaligned key is generally a bad idea, we just error loudly in dask.
+
+    For more information see pandas GH issue #15244 and Dask GH issue #1876."""
+    by = kwargs.get('by', None)
+    if by is not None and not _is_aligned(df, by):
+        msg = ("Grouping by an unaligned index is unsafe and unsupported.\n"
+               "This can be caused by filtering only one of the object or\n"
+               "grouping key. For example, the following works in pandas,\n"
+               "but not in dask:\n"
+               "\n"
+               "df[df.foo < 0].groupby(df.bar)\n"
+               "\n"
+               "This can be avoided by either filtering beforehand, or\n"
+               "passing in the name of the column instead:\n"
+               "\n"
+               "df2 = df[df.foo < 0]\n"
+               "df2.groupby(df2.bar)\n"
+               "# or\n"
+               "df[df.foo < 0].groupby('bar')\n"
+               "\n"
+               "For more information see dask GH issue #1876.")
+        raise ValueError(msg)
+    return df.groupby(**kwargs)
+
+
 def _groupby_slice_apply(df, grouper, key, func):
+    # No need to use raise if unaligned here - this is only called after
+    # shuffling, which makes everything aligned already
     g = df.groupby(grouper)
     if key:
         g = g[key]
@@ -35,7 +144,7 @@ def _groupby_slice_apply(df, grouper, key, func):
 
 def _groupby_get_group(df, by_key, get_key, columns):
     # SeriesGroupBy may pass df which includes group key
-    grouped = df.groupby(by_key)
+    grouped = _groupby_raise_unaligned(df, by=by_key)
 
     if get_key in grouped.groups:
         if isinstance(df, pd.DataFrame):
@@ -56,33 +165,39 @@ def _groupby_get_group(df, by_key, get_key, columns):
 ###############################################################
 
 def _groupby_aggregate(df, aggfunc=None, levels=None):
-    return aggfunc(df.groupby(level=levels))
+    return aggfunc(df.groupby(level=levels, sort=False))
 
 
-def _apply_chunk(df, index, func, columns):
+def _apply_chunk(df, *index, **kwargs):
+    func = kwargs.pop('chunk')
+    columns = kwargs.pop('columns')
+
+    g = _groupby_raise_unaligned(df, by=index)
+
     if isinstance(df, pd.Series) or columns is None:
-        return func(df.groupby(index))
+        return func(g)
     else:
-        columns = columns if isinstance(columns, str) else list(columns)
-        return func(df.groupby(index)[columns])
+        if isinstance(columns, (tuple, list, set, pd.Index)):
+            columns = list(columns)
+        return func(g[columns])
 
 
-def _var_chunk(df, index):
+def _var_chunk(df, *index):
     if isinstance(df, pd.Series):
         df = df.to_frame()
-    g = df.groupby(index)
+    g = _groupby_raise_unaligned(df, by=index)
     x = g.sum()
     x2 = g.agg(lambda x: (x**2).sum()).rename(columns=lambda c: c + '-x2')
     n = g.count().rename(columns=lambda c: c + '-count')
     return pd.concat([x, x2, n], axis=1)
 
 
-def _var_combine(g):
-    return g.groupby(level=0).sum()
+def _var_combine(g, levels):
+    return g.groupby(level=levels, sort=False).sum()
 
 
-def _var_agg(g, ddof):
-    g = g.groupby(level=0).sum()
+def _var_agg(g, levels, ddof):
+    g = g.groupby(level=levels, sort=False).sum()
     nc = len(g.columns)
     x = g[g.columns[:nc // 3]]
     x2 = g[g.columns[nc // 3:2 * nc // 3]].rename(columns=lambda c: c[:-3])
@@ -102,39 +217,49 @@ def _var_agg(g, ddof):
 # nunique
 ###############################################################
 
-def _nunique_df_chunk(df, index):
-    # we call set_index here to force a possibly duplicate index
+def _nunique_df_chunk(df, *index, **kwargs):
+    levels = kwargs.pop('levels')
+    name = kwargs.pop('name')
+
+    g = _groupby_raise_unaligned(df, by=index)
+    grouped = g[[name]].apply(pd.DataFrame.drop_duplicates)
+
+    # we set the index here to force a possibly duplicate index
     # for our reduce step
-    grouped = df.groupby(index).apply(pd.DataFrame.drop_duplicates)
-    grouped.index = grouped.index.get_level_values(level=0)
+    if isinstance(levels, list):
+        grouped.index = pd.MultiIndex.from_arrays([
+            grouped.index.get_level_values(level=level) for level in levels
+        ])
+    else:
+        grouped.index = grouped.index.get_level_values(level=levels)
+
     return grouped
 
 
-def _nunique_df_combine(df):
-    result = df.groupby(level=0).apply(pd.DataFrame.drop_duplicates)
-    result.index = result.index.get_level_values(level=0)
+def _nunique_df_combine(df, levels):
+    result = df.groupby(level=levels, sort=False).apply(pd.DataFrame.drop_duplicates)
+
+    if isinstance(levels, list):
+        result.index = pd.MultiIndex.from_arrays([
+            result.index.get_level_values(level=level) for level in levels
+        ])
+    else:
+        result.index = result.index.get_level_values(level=levels)
+
     return result
 
 
-def _nunique_df_aggregate(df, name):
-    return df.groupby(level=0)[name].nunique()
+def _nunique_df_aggregate(df, levels, name):
+    return df.groupby(level=levels, sort=False)[name].nunique()
 
 
-def _nunique_series_chunk(df, index):
+def _nunique_series_chunk(df, *index, **_ignored_):
+    # convert series to data frame, then hand over to dataframe code path
     assert isinstance(df, pd.Series)
-    if isinstance(index, np.ndarray):
-        assert len(index) == len(df)
-        index = pd.Series(index, index=df.index)
-    grouped = pd.concat([df, index], axis=1).drop_duplicates()
-    return grouped
 
-
-def _nunique_series_combine(df):
-    return df.drop_duplicates()
-
-
-def _nunique_series_aggregate(df):
-    return df.groupby(df.columns[1])[df.columns[0]].nunique()
+    df = df.to_frame()
+    kwargs = dict(name=df.columns[0], levels=_determine_levels(index))
+    return _nunique_df_chunk(df, *index, **kwargs)
 
 
 ###############################################################
@@ -399,7 +524,7 @@ def _groupby_apply_funcs(df, *index, **kwargs):
         kwargs.update(by=list(index))
 
     funcs = kwargs.pop('funcs')
-    grouped = df.groupby(**kwargs)
+    grouped = _groupby_raise_unaligned(df, **kwargs)
 
     result = collections.OrderedDict()
     for result_column, func, func_kwargs in funcs:
@@ -451,24 +576,20 @@ def _finalize_std(df, count_column, sum_column, sum2_column, ddof=1):
     return np.sqrt(result)
 
 
-def _normalize_index(df, index):
-    if not isinstance(df, DataFrame):
-        return index
+def _cum_agg_aligned(part, cum_last, index, columns, func, initial):
+    align = cum_last.reindex(part.set_index(index).index, fill_value=initial)
+    align.index = part.index
+    return func(part[columns], align)
 
-    elif isinstance(index, list):
-        return [_normalize_index(df, col) for col in index]
 
-    elif (isinstance(index, Series) and index.name in df.columns and
-          index._name == df[index.name]._name):
-            return index.name
+def _cum_agg_filled(a, b, func, initial):
+    union = a.index.union(b.index)
+    return func(a.reindex(union, fill_value=initial),
+                b.reindex(union, fill_value=initial), fill_value=initial)
 
-    elif (isinstance(index, DataFrame) and
-          set(index.columns).issubset(df.columns) and
-          index._name == df[index.columns]._name):
-        return list(index.columns)
 
-    else:
-        return index
+def _cumcount_aggregate(a, b, fill_value=None):
+    return a.add(b, fill_value=fill_value) + 1
 
 
 class _GroupBy(object):
@@ -479,58 +600,45 @@ class _GroupBy(object):
 
     obj: DataFrame or Series
         DataFrame or Series to be grouped
-    index: str, list or Series
+    by: str, list or Series
         The key for grouping
-    kwargs: dict
-        Other keywords passed to groupby
+    slice: str, list
+        The slice keys applied to GroupBy result
     """
-    def __init__(self, df, index=None, slice=None, **kwargs):
+    def __init__(self, df, by=None, slice=None):
+
         assert isinstance(df, (DataFrame, Series))
         self.obj = df
-
         # grouping key passed via groupby method
-        self.index = _normalize_index(df, index)
+        self.index = _normalize_index(df, by)
+
+        if isinstance(self.index, list):
+            do_index_partition_align = all(
+                item.divisions == df.divisions if isinstance(item, Series) else True
+                for item in self.index
+            )
+        elif isinstance(self.index, Series):
+            do_index_partition_align = df.divisions == self.index.divisions
+        else:
+            do_index_partition_align = True
+
+        if not do_index_partition_align:
+            raise NotImplementedError("The grouped object and index of the "
+                                      "groupby must have the same divisions.")
 
         # slicing key applied to _GroupBy instance
         self._slice = slice
 
-        self.kwargs = kwargs
-
-        if isinstance(index, Series) and df.divisions != index.divisions:
-            msg = ("The Series and index of the groupby"
-                   " must have the same divisions.")
-            raise NotImplementedError(msg)
-
-        if self._is_grouped_by_sliced_column(self.obj, index):
-            # check whether given Series is taken from given df and unchanged.
-            # If any operations are performed, _name will be changed to
-            # e.g. "elemwise-xxxx"
-
-            # if group key (index) is a Series sliced from DataFrame,
-            # emulation must be performed as the same.
-            # otherwise, group key is regarded as a separate column
-            self._meta = self.obj._meta.groupby(self.obj._meta[index.name])
+        if isinstance(self.index, list):
+            index_meta = [item._meta if isinstance(item, Series) else item for item in self.index]
 
         elif isinstance(self.index, Series):
-            self._meta = self.obj._meta.groupby(self.index._meta)
-        else:
-            self._meta = self.obj._meta.groupby(self.index)
+            index_meta = self.index._meta
 
-    def _is_grouped_by_sliced_column(self, df, index):
-        """
-        Return whether index is a Series sliced from df
-        """
-        if isinstance(df, Series):
-            return False
-        if (isinstance(index, Series) and index._name in df.columns and
-                index._name == df[index.name]._name):
-            return True
-        if (isinstance(index, DataFrame) and
-                set(index.columns).issubset(df.columns) and
-                index._name == df[index.columns]._name):
-            index = list(index.columns)
-            return True
-        return False
+        else:
+            index_meta = self.index
+
+        self._meta = self.obj._meta.groupby(index_meta)
 
     @property
     def _meta_nonempty(self):
@@ -538,16 +646,21 @@ class _GroupBy(object):
         Return a pd.DataFrameGroupBy / pd.SeriesGroupBy which contains sample data.
         """
         sample = self.obj._meta_nonempty
-        if isinstance(self.index, Series):
-            if self._is_grouped_by_sliced_column(self.obj, self.index):
-                grouped = sample.groupby(sample[self.index.name])
-            else:
-                grouped = sample.groupby(self.index._meta_nonempty)
+
+        if isinstance(self.index, list):
+            index_meta = [item._meta_nonempty if isinstance(item, Series) else item for item in self.index]
+
+        elif isinstance(self.index, Series):
+            index_meta = self.index._meta_nonempty
+
         else:
-            grouped = sample.groupby(self.index)
+            index_meta = self.index
+
+        grouped = sample.groupby(index_meta)
         return _maybe_slice(grouped, self._slice)
 
-    def _aca_agg(self, token, func, aggfunc=None, split_every=None):
+    def _aca_agg(self, token, func, aggfunc=None, split_every=None,
+                 split_out=1):
         if aggfunc is None:
             aggfunc = func
 
@@ -555,49 +668,134 @@ class _GroupBy(object):
         columns = meta.name if isinstance(meta, pd.Series) else meta.columns
 
         token = self._token_prefix + token
+        levels = _determine_levels(self.index)
 
-        if isinstance(self.index, (tuple, list)) and len(self.index) > 1:
-            levels = list(range(len(self.index)))
-        else:
-            levels = 0
-
-        return aca([self.obj, self.index, func, columns],
-                   chunk=_apply_chunk, aggregate=_groupby_aggregate,
+        return aca([self.obj, self.index] if not isinstance(self.index, list) else [self.obj] + self.index,
+                   chunk=_apply_chunk,
+                   chunk_kwargs=dict(chunk=func, columns=columns),
+                   aggregate=_groupby_aggregate,
                    meta=meta, token=token, split_every=split_every,
-                   aggregate_kwargs=dict(aggfunc=aggfunc, levels=levels))
+                   aggregate_kwargs=dict(aggfunc=aggfunc, levels=levels),
+                   split_out=split_out, split_out_setup=split_out_on_index)
+
+    def _cum_agg(self, token, chunk, aggregate, initial):
+        """ Wrapper for cumulative groupby operation """
+        meta = chunk(self._meta)
+        columns = meta.name if isinstance(meta, pd.Series) else meta.columns
+        index = self.index if isinstance(self.index, list) else [self.index]
+
+        name = self._token_prefix + token
+        name_part = name + '-map'
+        name_last = name + '-take-last'
+        name_cum = name + '-cum-last'
+
+        # cumulate each partitions
+        cumpart_raw = map_partitions(_apply_chunk, self.obj, *index,
+                                     chunk=chunk,
+                                     columns=columns,
+                                     token=name_part,
+                                     meta=meta)
+        cumpart_ext = (cumpart_raw.to_frame()
+                       if isinstance(meta, pd.Series)
+                       else cumpart_raw).assign(**{i: self.obj[i]
+                                                   for i in index})
+
+        cumlast = map_partitions(_apply_chunk, cumpart_ext, *index,
+                                 columns=0 if columns is None else columns,
+                                 chunk=M.last,
+                                 meta=meta,
+                                 token=name_last)
+
+        # aggregate cumulated partisions and its previous last element
+        dask = {}
+        dask[(name, 0)] = (cumpart_raw._name, 0)
+
+        for i in range(1, self.obj.npartitions):
+            # store each cumulative step to graph to reduce computation
+            if i == 1:
+                dask[(name_cum, i)] = (cumlast._name, i - 1)
+            else:
+                # aggregate with previous cumulation results
+                dask[(name_cum, i)] = (_cum_agg_filled,
+                                       (name_cum, i - 1),
+                                       (cumlast._name, i - 1),
+                                       aggregate, initial)
+            dask[(name, i)] = (_cum_agg_aligned,
+                               (cumpart_ext._name, i), (name_cum, i),
+                               index, 0 if columns is None else columns,
+                               aggregate, initial)
+        return new_dd_object(merge(dask, cumpart_ext.dask, cumlast.dask),
+                             name, chunk(self._meta), self.obj.divisions)
 
     @derived_from(pd.core.groupby.GroupBy)
-    def sum(self, split_every=None):
-        return self._aca_agg(token='sum', func=M.sum, split_every=split_every)
+    def cumsum(self, axis=0):
+        if axis:
+            return self.obj.cumsum(axis=axis)
+        else:
+            return self._cum_agg('cumsum',
+                                 chunk=M.cumsum,
+                                 aggregate=M.add,
+                                 initial=0)
 
     @derived_from(pd.core.groupby.GroupBy)
-    def min(self, split_every=None):
-        return self._aca_agg(token='min', func=M.min, split_every=split_every)
+    def cumprod(self, axis=0):
+        if axis:
+            return self.obj.cumprod(axis=axis)
+        else:
+            return self._cum_agg('cumprod',
+                                 chunk=M.cumprod,
+                                 aggregate=M.mul,
+                                 initial=1)
 
     @derived_from(pd.core.groupby.GroupBy)
-    def max(self, split_every=None):
-        return self._aca_agg(token='max', func=M.max, split_every=split_every)
+    def cumcount(self, axis=None):
+        return self._cum_agg('cumcount',
+                             chunk=M.cumcount,
+                             aggregate=_cumcount_aggregate,
+                             initial=-1)
 
     @derived_from(pd.core.groupby.GroupBy)
-    def count(self, split_every=None):
+    def sum(self, split_every=None, split_out=1):
+        return self._aca_agg(token='sum', func=M.sum, split_every=split_every,
+                             split_out=split_out)
+
+    @derived_from(pd.core.groupby.GroupBy)
+    def min(self, split_every=None, split_out=1):
+        return self._aca_agg(token='min', func=M.min, split_every=split_every,
+                             split_out=split_out)
+
+    @derived_from(pd.core.groupby.GroupBy)
+    def max(self, split_every=None, split_out=1):
+        return self._aca_agg(token='max', func=M.max, split_every=split_every,
+                             split_out=split_out)
+
+    @derived_from(pd.core.groupby.GroupBy)
+    def count(self, split_every=None, split_out=1):
         return self._aca_agg(token='count', func=M.count,
-                             aggfunc=M.sum, split_every=split_every)
+                             aggfunc=M.sum, split_every=split_every,
+                             split_out=split_out)
 
     @derived_from(pd.core.groupby.GroupBy)
-    def mean(self, split_every=None):
-        return self.sum(split_every=split_every) / self.count(split_every=split_every)
+    def mean(self, split_every=None, split_out=1):
+        return (self.sum(split_every=split_every, split_out=split_out) /
+                self.count(split_every=split_every, split_out=split_out))
 
     @derived_from(pd.core.groupby.GroupBy)
-    def size(self, split_every=None):
+    def size(self, split_every=None, split_out=1):
         return self._aca_agg(token='size', func=M.size, aggfunc=M.sum,
-                             split_every=split_every)
+                             split_every=split_every, split_out=split_out)
 
     @derived_from(pd.core.groupby.GroupBy)
-    def var(self, ddof=1, split_every=None):
-        result = aca([self.obj, self.index], chunk=_var_chunk,
+    def var(self, ddof=1, split_every=None, split_out=1):
+        levels = _determine_levels(self.index)
+        result = aca([self.obj, self.index] if not isinstance(self.index, list) else [self.obj] + self.index,
+                     chunk=_var_chunk,
                      aggregate=_var_agg, combine=_var_combine,
                      token=self._token_prefix + 'var',
-                     aggregate_kwargs={'ddof': ddof}, split_every=split_every)
+                     aggregate_kwargs={'ddof': ddof, 'levels': levels},
+                     combine_kwargs={'levels': levels},
+                     split_every=split_every, split_out=split_out,
+                     split_out_setup=split_out_on_index)
 
         if isinstance(self.obj, Series):
             result = result[result.columns[0]]
@@ -607,8 +805,8 @@ class _GroupBy(object):
         return result
 
     @derived_from(pd.core.groupby.GroupBy)
-    def std(self, ddof=1, split_every=None):
-        v = self.var(ddof, split_every=split_every)
+    def std(self, ddof=1, split_every=None, split_out=1):
+        v = self.var(ddof, split_every=split_every, split_out=split_out)
         result = map_partitions(np.sqrt, v, meta=v)
         return result
 
@@ -624,7 +822,7 @@ class _GroupBy(object):
         return map_partitions(_groupby_get_group, self.obj, self.index, key,
                               columns, meta=meta, token=token)
 
-    def aggregate(self, arg, split_every):
+    def aggregate(self, arg, split_every, split_out=1):
         if isinstance(self.obj, DataFrame):
             if isinstance(self.index, tuple) or np.isscalar(self.index):
                 group_columns = {self.index}
@@ -636,10 +834,17 @@ class _GroupBy(object):
             else:
                 group_columns = set()
 
-            # NOTE: this step relies on the index normalization to replace
-            #       series with their name in an index.
-            non_group_columns = [col for col in self.obj.columns
-                                 if col not in group_columns]
+            if self._slice:
+                # pandas doesn't exclude the grouping column in a SeriesGroupBy
+                # like df.groupby('a')['a'].agg(...)
+                non_group_columns = self._slice
+                if not isinstance(non_group_columns, list):
+                    non_group_columns = [non_group_columns]
+            else:
+                # NOTE: this step relies on the index normalization to replace
+                #       series with their name in an index.
+                non_group_columns = [col for col in self.obj.columns
+                                     if col not in group_columns]
 
             spec = _normalize_spec(arg, non_group_columns)
 
@@ -661,14 +866,6 @@ class _GroupBy(object):
         else:
             levels = 0
 
-        # apply the transformations to determine the meta object
-        meta_groupby = pd.Series([], dtype=bool, index=self.obj._meta.index)
-        meta_stage1 = _groupby_apply_funcs(self.obj._meta, funcs=chunk_funcs,
-                                           by=meta_groupby)
-        meta_stage2 = _groupby_apply_funcs(meta_stage1, funcs=aggregate_funcs,
-                                           level=0)
-        meta = _agg_finalize(meta_stage2, finalizers)
-
         if not isinstance(self.index, list):
             chunk_args = [self.obj, self.index]
 
@@ -682,13 +879,14 @@ class _GroupBy(object):
                   aggregate_kwargs=dict(funcs=aggregate_funcs, level=levels),
                   combine=_groupby_apply_funcs,
                   combine_kwargs=dict(funcs=aggregate_funcs, level=levels),
-                  meta=meta, token='aggregate', split_every=split_every)
+                  token='aggregate', split_every=split_every,
+                  split_out=split_out, split_out_setup=split_out_on_index)
 
-        return map_partitions(_agg_finalize, obj, meta=meta,
-                              token='aggregate-finalize', funcs=finalizers)
+        return map_partitions(_agg_finalize, obj, token='aggregate-finalize',
+                              funcs=finalizers)
 
     @insert_meta_param_description(pad=12)
-    def apply(self, func, meta=no_default, columns=no_default):
+    def apply(self, func, meta=no_default):
         """ Parallel version of pandas GroupBy.apply
 
         This mimics the pandas version except for the following:
@@ -702,29 +900,18 @@ class _GroupBy(object):
         func: function
             Function to apply
         $META
-        columns: list, scalar or None
-            Deprecated, use `meta` instead. If list is given, the result is a
-            DataFrame which columns is specified list. Otherwise, the result is
-            a Series which name is given scalar or None (no name). If name
-            keyword is not given, dask tries to infer the result type using its
-            beginning of data. This inference may take some time and lead to
-            unexpected result
 
         Returns
         -------
         applied : Series or DataFrame depending on columns keyword
         """
-        if columns is not no_default:
-            warnings.warn("`columns` is deprecated, please use `meta` instead")
-            if meta is no_default and isinstance(columns, (pd.DataFrame, pd.Series)):
-                meta = columns
         if meta is no_default:
             msg = ("`meta` is not specified, inferred from partial data. "
                    "Please provide `meta` if the result is unexpected.\n"
                    "  Before: .apply(func)\n"
                    "  After:  .apply(func, meta={'x': 'f8', 'y': 'f8'}) for dataframe result\n"
                    "  or:     .apply(func, meta=('x', 'f8'))            for series result")
-            warnings.warn(msg)
+            warnings.warn(msg, stacklevel=2)
 
             with raise_on_meta_error("groupby.apply({0})".format(funcname(func))):
                 meta = self._meta_nonempty.apply(func)
@@ -738,22 +925,31 @@ class _GroupBy(object):
         elif isinstance(self.index, Series):
             df2 = df.assign(_index=self.index)
             index = self.index
+        elif (isinstance(self.index, list) and
+              any(isinstance(item, Series) for item in self.index)):
+            raise NotImplementedError("groupby-apply with a multiple Series "
+                                      "is currently not supported")
         else:
             df2 = df
             index = df[self.index]
 
-        df3 = shuffle(df2, index, **self.kwargs)  # shuffle dataframe and index
+        df3 = shuffle(df2, index)  # shuffle dataframe and index
 
         if isinstance(self.index, DataFrame):  # extract index from dataframe
             cols = ['_index_' + c for c in self.index.columns]
             index2 = df3[cols]
-            df4 = df3.drop(cols, axis=1, dtype=meta.columns.dtype if
-                           isinstance(meta, pd.DataFrame) else None)
+            if isinstance(meta, pd.DataFrame):
+                df4 = df3.map_partitions(drop_columns, cols, meta.columns.dtype)
+            else:
+                df4 = df3.drop(cols, axis=1)
         elif isinstance(self.index, Series):
             index2 = df3['_index']
             index2.name = self.index.name
-            df4 = df3.drop('_index', axis=1, dtype=meta.columns.dtype if
-                           isinstance(meta, DataFrame) else None)
+            if isinstance(meta, pd.DataFrame):
+                df4 = df3.map_partitions(drop_columns, '_index',
+                                         meta.columns.dtype)
+            else:
+                df4 = df3.drop('_index', axis=1)
         else:
             df4 = df3
             index2 = self.index
@@ -769,23 +965,11 @@ class DataFrameGroupBy(_GroupBy):
 
     _token_prefix = 'dataframe-groupby-'
 
-    def __init__(self, df, index=None, slice=None, **kwargs):
-
-        if not kwargs.get('as_index', True):
-            msg = ("The keyword argument `as_index=False` is not supported in "
-                   "dask.dataframe.groupby")
-            raise NotImplementedError(msg)
-
-        super(DataFrameGroupBy, self).__init__(df, index=index,
-                                               slice=slice, **kwargs)
-
     def __getitem__(self, key):
         if isinstance(key, list):
-            g = DataFrameGroupBy(self.obj, index=self.index,
-                                 slice=key, **self.kwargs)
+            g = DataFrameGroupBy(self.obj, by=self.index, slice=key)
         else:
-            g = SeriesGroupBy(self.obj, index=self.index,
-                              slice=key, **self.kwargs)
+            g = SeriesGroupBy(self.obj, by=self.index, slice=key)
 
         # error is raised from pandas
         g._meta = g._meta[key]
@@ -802,70 +986,77 @@ class DataFrameGroupBy(_GroupBy):
             raise AttributeError(e)
 
     @derived_from(pd.core.groupby.DataFrameGroupBy)
-    def aggregate(self, arg, split_every=None):
+    def aggregate(self, arg, split_every=None, split_out=1):
         if arg == 'size':
             return self.size()
 
-        return super(DataFrameGroupBy, self).aggregate(arg, split_every=split_every)
+        return super(DataFrameGroupBy, self).aggregate(arg, split_every=split_every, split_out=split_out)
 
     @derived_from(pd.core.groupby.DataFrameGroupBy)
-    def agg(self, arg, split_every=None):
-        return self.aggregate(arg, split_every=split_every)
+    def agg(self, arg, split_every=None, split_out=1):
+        return self.aggregate(arg, split_every=split_every, split_out=split_out)
 
 
 class SeriesGroupBy(_GroupBy):
 
     _token_prefix = 'series-groupby-'
 
-    def __init__(self, df, index, slice=None, **kwargs):
+    def __init__(self, df, by=None, slice=None):
+        # for any non series object, raise pandas-compat error message
 
-        # raise pandas-compat error message
         if isinstance(df, Series):
-            # When obj is Series, index must be Series
-            if not isinstance(index, Series):
-                if isinstance(index, list):
-                    if len(index) == 0:
-                        raise ValueError("No group keys passed!")
-                    msg = "Grouper for '{0}' not 1-dimensional"
-                    raise ValueError(msg.format(index[0]))
-                # raise error from pandas
-                df._meta.groupby(index)
-        super(SeriesGroupBy, self).__init__(df, index=index,
-                                            slice=slice, **kwargs)
+            if isinstance(by, Series):
+                pass
+            elif isinstance(by, list):
+                if len(by) == 0:
+                    raise ValueError("No group keys passed!")
 
-    def nunique(self, split_every=None):
+                non_series_items = [item for item in by
+                                    if not isinstance(item, Series)]
+                # raise error from pandas, if applicable
+                df._meta.groupby(non_series_items)
+            else:
+                # raise error from pandas, if applicable
+                df._meta.groupby(by)
+
+        super(SeriesGroupBy, self).__init__(df, by=by, slice=slice)
+
+    def nunique(self, split_every=None, split_out=1):
         name = self._meta.obj.name
-        meta = pd.Series([], dtype='int64',
-                         index=pd.Index([], dtype=self._meta.obj.dtype),
-                         name=name)
+        levels = _determine_levels(self.index)
 
         if isinstance(self.obj, DataFrame):
-            return aca([self.obj, self.index],
-                       chunk=_nunique_df_chunk,
-                       aggregate=_nunique_df_aggregate,
-                       combine=_nunique_df_combine,
-                       meta=meta, token='series-groupby-nunique',
-                       aggregate_kwargs={'name': name},
-                       split_every=split_every)
+            chunk = _nunique_df_chunk
+
         else:
-            return aca([self.obj, self.index],
-                       chunk=_nunique_series_chunk,
-                       aggregate=_nunique_series_aggregate,
-                       combine=_nunique_series_combine,
-                       meta=meta, token='series-groupby-nunique',
-                       split_every=split_every)
+            chunk = _nunique_series_chunk
+
+        return aca([self.obj, self.index] if not isinstance(self.index, list) else [self.obj] + self.index,
+                   chunk=chunk,
+                   aggregate=_nunique_df_aggregate,
+                   combine=_nunique_df_combine,
+                   token='series-groupby-nunique',
+                   chunk_kwargs={'levels': levels, 'name': name},
+                   aggregate_kwargs={'levels': levels, 'name': name},
+                   combine_kwargs={'levels': levels},
+                   split_every=split_every, split_out=split_out,
+                   split_out_setup=split_out_on_index)
 
     @derived_from(pd.core.groupby.SeriesGroupBy)
-    def aggregate(self, arg, split_every=None):
+    def aggregate(self, arg, split_every=None, split_out=1):
         # short-circuit 'simple' aggregations
         if (
             not isinstance(arg, (list, dict)) and
             arg in {'sum', 'mean', 'var', 'size', 'std', 'count'}
         ):
-            return getattr(self, arg)(split_every=split_every)
+            return getattr(self, arg)(split_every=split_every,
+                                      split_out=split_out)
 
-        return super(SeriesGroupBy, self).aggregate(arg, split_every=split_every)
+        result = super(SeriesGroupBy, self).aggregate(arg, split_every=split_every, split_out=split_out)
+        if self._slice:
+            result = result[self._slice]
+        return result
 
     @derived_from(pd.core.groupby.SeriesGroupBy)
-    def agg(self, arg, split_every=None):
-        return self.aggregate(arg, split_every=split_every)
+    def agg(self, arg, split_every=None, split_out=1):
+        return self.aggregate(arg, split_every=split_every, split_out=split_out)
