@@ -3,7 +3,6 @@ import pandas as pd
 import pandas.util.testing as tm
 
 import pytest
-from distutils.version import LooseVersion
 from threading import Lock
 
 import threading
@@ -11,10 +10,10 @@ import threading
 import dask.array as da
 import dask.dataframe as dd
 from dask.dataframe.io.io import _meta_from_array
-from dask.delayed import Delayed
+from dask.delayed import Delayed, delayed
 
 from dask.utils import tmpfile
-from dask.async import get_sync
+from dask.local import get_sync
 
 from dask.dataframe.utils import assert_eq
 
@@ -232,10 +231,12 @@ def test_from_pandas_dataframe():
     tm.assert_frame_equal(df, ddf.compute())
     ddf = dd.from_pandas(df, chunksize=8)
     msg = 'Exactly one of npartitions and chunksize must be specified.'
-    with tm.assertRaisesRegexp(ValueError, msg):
+    with pytest.raises(ValueError) as err:
         dd.from_pandas(df, npartitions=2, chunksize=2)
-    with tm.assertRaisesRegexp((ValueError, AssertionError), msg):
+    assert msg in str(err.value)
+    with pytest.raises((ValueError, AssertionError)) as err:
         dd.from_pandas(df)
+    assert msg in str(err.value)
     assert len(ddf.dask) == 3
     assert len(ddf.divisions) == len(ddf.dask) + 1
     assert isinstance(ddf.divisions[0], type(df.index[0]))
@@ -309,12 +310,14 @@ def test_from_pandas_single_row():
     assert_eq(ddf, df)
 
 
+@pytest.mark.skipif(np.__version__ < '1.11',
+                    reason='datetime unit unsupported in NumPy < 1.11')
 def test_from_pandas_with_datetime_index():
     df = pd.DataFrame({"Date": ["2015-08-28", "2015-08-27", "2015-08-26",
                                 "2015-08-25", "2015-08-24", "2015-08-21",
                                 "2015-08-20", "2015-08-19", "2015-08-18"],
                        "Val": list(range(9))})
-    df.Date = df.Date.astype('datetime64')
+    df.Date = df.Date.astype('datetime64[ns]')
     ddf = dd.from_pandas(df, 2)
     assert_eq(df, ddf)
     ddf = dd.from_pandas(df, chunksize=2)
@@ -439,98 +442,28 @@ def test_from_dask_array_struct_dtype():
               pd.DataFrame(x, columns=['b', 'a']))
 
 
-def test_to_castra():
-    castra = pytest.importorskip('castra')
-    blosc = pytest.importorskip('blosc')
-    if (LooseVersion(blosc.__version__) == '1.3.0' or
-            LooseVersion(castra.__version__) < '0.1.8'):
-        pytest.skip()
-    df = pd.DataFrame({'x': ['a', 'b', 'c', 'd'],
-                       'y': [2, 3, 4, 5]},
-                      index=pd.Index([1., 2., 3., 4.], name='ind'))
-    a = dd.from_pandas(df, 2)
+def test_from_dask_array_unknown_chunks():
+    # Series
+    dx = da.Array({('x', 0): np.arange(5), ('x', 1): np.arange(5, 11)}, 'x',
+                  ((np.nan, np.nan,),), np.arange(1).dtype)
+    df = dd.from_dask_array(dx)
+    assert isinstance(df, dd.Series)
+    assert not df.known_divisions
+    assert_eq(df, pd.Series(np.arange(11)), check_index=False)
 
-    c = a.to_castra()
-    b = c.to_dask()
-    try:
-        tm.assert_frame_equal(df, c[:])
-        tm.assert_frame_equal(b.compute(), df)
-    finally:
-        c.drop()
+    # DataFrame
+    dsk = {('x', 0, 0): np.random.random((2, 3)),
+           ('x', 1, 0): np.random.random((5, 3))}
+    dx = da.Array(dsk, 'x', ((np.nan, np.nan,), (3,)), np.float64)
+    df = dd.from_dask_array(dx)
+    assert isinstance(df, dd.DataFrame)
+    assert not df.known_divisions
+    assert_eq(df, pd.DataFrame(dx.compute()), check_index=False)
 
-    c = a.to_castra(categories=['x'])
-    try:
-        assert c[:].dtypes['x'] == 'category'
-    finally:
-        c.drop()
-
-    c = a.to_castra(sorted_index_column='y')
-    try:
-        tm.assert_frame_equal(c[:], df.set_index('y'))
-    finally:
-        c.drop()
-
-    delayed = a.to_castra(compute=False)
-    assert isinstance(delayed, Delayed)
-    c = delayed.compute()
-    try:
-        tm.assert_frame_equal(c[:], df)
-    finally:
-        c.drop()
-
-    # make sure compute=False preserves the same interface
-    c1 = a.to_castra(compute=True)
-    c2 = a.to_castra(compute=False).compute()
-    try:
-        tm.assert_frame_equal(c1[:], c2[:])
-    finally:
-        c1.drop()
-        c2.drop()
-
-
-def test_from_castra():
-    castra = pytest.importorskip('castra')
-    blosc = pytest.importorskip('blosc')
-    if (LooseVersion(blosc.__version__) == '1.3.0' or
-            LooseVersion(castra.__version__) < '0.1.8'):
-        pytest.skip()
-    df = pd.DataFrame({'x': ['a', 'b', 'c', 'd'],
-                       'y': [2, 3, 4, 5]},
-                      index=pd.Index([1., 2., 3., 4.], name='ind'))
-    a = dd.from_pandas(df, 2)
-
-    c = a.to_castra()
-    with_castra = dd.from_castra(c)
-    with_fn = dd.from_castra(c.path)
-    with_columns = dd.from_castra(c, 'x')
-    try:
-        tm.assert_frame_equal(df, with_castra.compute())
-        tm.assert_frame_equal(df, with_fn.compute())
-        tm.assert_series_equal(df.x, with_columns.compute())
-    finally:
-        # Calling c.drop() is a race condition on drop from `with_fn.__del__`
-        # and c.drop. Manually `del`ing gets around this.
-        del with_fn, c
-
-
-def test_from_castra_with_selection():
-    """ Optimizations fuse getitems with load_partitions
-
-    We used to use getitem for both column access and selections
-    """
-    castra = pytest.importorskip('castra')
-    blosc = pytest.importorskip('blosc')
-    if (LooseVersion(blosc.__version__) == '1.3.0' or
-            LooseVersion(castra.__version__) < '0.1.8'):
-        pytest.skip()
-    df = pd.DataFrame({'x': ['a', 'b', 'c', 'd'],
-                       'y': [2, 3, 4, 5]},
-                      index=pd.Index([1., 2., 3., 4.], name='ind'))
-    a = dd.from_pandas(df, 2)
-
-    b = dd.from_castra(a.to_castra())
-
-    assert_eq(b[b.y > 3].x, df[df.y > 3].x)
+    # Unknown width
+    dx = da.Array(dsk, 'x', ((np.nan, np.nan,), (np.nan,)), np.float64)
+    with pytest.raises(ValueError):
+        df = dd.from_dask_array(dx)
 
 
 def test_to_bag():
@@ -544,3 +477,67 @@ def test_to_bag():
     assert ddf.to_bag(True).compute() == list(a.itertuples(True))
     assert ddf.x.to_bag(True).compute() == list(a.x.iteritems())
     assert ddf.x.to_bag().compute() == list(a.x)
+
+
+def test_to_records():
+    pytest.importorskip('dask.array')
+    from dask.array.utils import assert_eq
+    df = pd.DataFrame({'x': ['a', 'b', 'c', 'd'],
+                       'y': [2, 3, 4, 5]},
+                      index=pd.Index([1., 2., 3., 4.], name='ind'))
+    ddf = dd.from_pandas(df, 2)
+
+    assert_eq(df.to_records(), ddf.to_records())
+
+
+def test_from_delayed():
+    df = pd.DataFrame(data=np.random.normal(size=(10, 4)), columns=list('abcd'))
+    parts = [df.iloc[:1], df.iloc[1:3], df.iloc[3:6], df.iloc[6:10]]
+    dfs = [delayed(parts.__getitem__)(i) for i in range(4)]
+    meta = dfs[0].compute()
+
+    my_len = lambda x: pd.Series([len(x)])
+
+    for divisions in [None, [0, 1, 3, 6, 10]]:
+        ddf = dd.from_delayed(dfs, meta=meta, divisions=divisions)
+        assert_eq(ddf, df)
+        assert list(ddf.map_partitions(my_len).compute()) == [1, 2, 3, 4]
+        assert ddf.known_divisions == (divisions is not None)
+
+        s = dd.from_delayed([d.a for d in dfs], meta=meta.a,
+                            divisions=divisions)
+        assert_eq(s, df.a)
+        assert list(s.map_partitions(my_len).compute()) == [1, 2, 3, 4]
+        assert ddf.known_divisions == (divisions is not None)
+
+    with pytest.raises(ValueError):
+        dd.from_delayed(dfs, meta=meta, divisions=[0, 1, 3, 6])
+
+
+def test_from_delayed_sorted():
+    a = pd.DataFrame({'x': [1, 2]}, index=[1, 10])
+    b = pd.DataFrame({'x': [4, 1]}, index=[100, 200])
+
+    A = dd.from_delayed([delayed(a), delayed(b)], divisions='sorted')
+    assert A.known_divisions
+
+    assert A.divisions == (1, 100, 200)
+
+
+def test_to_delayed():
+    df = pd.DataFrame({'x': [1, 2, 3, 4], 'y': [10, 20, 30, 40]})
+    ddf = dd.from_pandas(df, npartitions=2)
+    a, b = ddf.to_delayed()
+    assert isinstance(a, Delayed)
+    assert isinstance(b, Delayed)
+
+    assert_eq(a.compute(), df.iloc[:2])
+
+
+def test_to_delayed_optimizes():
+    df = pd.DataFrame({'x': list(range(20))})
+    ddf = dd.from_pandas(df, npartitions=20)
+    x = (ddf + 1).loc[:2]
+
+    d = x.to_delayed()[0]
+    assert len(d.dask) < 20
