@@ -6,25 +6,29 @@ import uuid
 
 import numpy as np
 import pandas as pd
-from pandas.core.categorical import is_categorical_dtype
 from toolz import merge
 
-from .core import DataFrame, Series, _Frame, _concat
+from .methods import drop_columns
+from .core import DataFrame, Series, _Frame, _concat, map_partitions
+from .hashing import hash_pandas_object
+from .utils import PANDAS_VERSION
 
-from ..base import tokenize
+from .. import base
+from ..base import tokenize, compute
 from ..context import _globals
+from ..delayed import delayed
+from ..sizeof import sizeof
 from ..utils import digit, insert, M
 
+if PANDAS_VERSION >= '0.20.0':
+    from pandas._libs.algos import groupsort_indexer
+else:
+    from pandas.algos import groupsort_indexer
 
-def set_index(df, index, npartitions=None, shuffle=None, compute=True,
-              drop=True, upsample=1.0, **kwargs):
-    """ Set DataFrame index to new column
 
-    Sorts index and realigns Dataframe to new sorted order.
-
-    This shuffles and repartitions your data. If done in parallel the
-    resulting order is non-deterministic.
-    """
+def set_index(df, index, npartitions=None, shuffle=None, compute=False,
+              drop=True, upsample=1.0, divisions=None, **kwargs):
+    """ See _Frame.set_index for docstring """
     if (isinstance(index, Series) and index._name == df.index._name):
         return df
     if isinstance(index, (DataFrame, tuple, list)):
@@ -33,18 +37,78 @@ def set_index(df, index, npartitions=None, shuffle=None, compute=True,
             "You tried to index with this index: %s\n"
             "Indexes must be single columns only." % str(index))
 
-    npartitions = npartitions or df.npartitions
+    if npartitions == 'auto':
+        repartition = True
+        npartitions = max(100, df.npartitions)
+    else:
+        if npartitions is None:
+            npartitions = df.npartitions
+        repartition = False
+
     if not isinstance(index, Series):
         index2 = df[index]
     else:
         index2 = index
 
-    divisions = (index2
-                 ._repartition_quantiles(npartitions, upsample=upsample)
-                 .compute()).tolist()
+    if divisions is None:
+        divisions = index2._repartition_quantiles(npartitions, upsample=upsample)
+        if repartition:
+            parts = df.to_delayed()
+            sizes = [delayed(sizeof)(part) for part in parts]
+        else:
+            sizes = []
+        iparts = index2.to_delayed()
+        mins = [ipart.min() for ipart in iparts]
+        maxes = [ipart.max() for ipart in iparts]
+        divisions, sizes, mins, maxes = base.compute(divisions, sizes, mins, maxes)
+        divisions = divisions.tolist()
+
+        if repartition:
+            total = sum(sizes)
+            npartitions = max(math.ceil(total / 128e6), 1)
+            npartitions = min(npartitions, df.npartitions)
+            n = len(divisions)
+            try:
+                divisions = np.interp(x=np.linspace(0, n - 1, npartitions + 1),
+                                      xp=np.linspace(0, n - 1, n),
+                                      fp=divisions).tolist()
+            except (TypeError, ValueError):  # str type
+                indexes = np.linspace(0, n - 1, npartitions + 1).astype(int)
+                divisions = [divisions[i] for i in indexes]
+
+        mins = remove_nans(mins)
+        maxes = remove_nans(maxes)
+
+        if (mins == sorted(mins) and maxes == sorted(maxes) and
+                all(mx < mn for mx, mn in zip(maxes[:-1], mins[1:]))):
+            divisions = mins + [maxes[-1]]
+            result = set_sorted_index(df, index, drop=drop, divisions=divisions)
+            # There are cases where this still may not be sorted
+            # so sort_index to be sure. https://github.com/dask/dask/issues/2288
+            return result.map_partitions(M.sort_index)
 
     return set_partition(df, index, divisions, shuffle=shuffle, drop=drop,
-                         **kwargs)
+                         compute=compute, **kwargs)
+
+
+def remove_nans(divisions):
+    """ Remove nans from divisions
+
+    These sometime pop up when we call min/max on an empty partition
+
+    Examples
+    --------
+    >>> remove_nans((np.nan, 1, 2))
+    [1, 1, 2]
+    >>> remove_nans((1, np.nan, 2))
+    [1, 2, 2]
+    """
+    divisions = list(divisions)
+    for i in range(len(divisions) - 2, -1, -1):
+        d = divisions[i]
+        if isinstance(d, float) and math.isnan(d):
+            divisions[i] = divisions[i + 1]
+    return divisions
 
 
 def set_partition(df, index, divisions, max_branch=32, drop=True, shuffle=None,
@@ -135,8 +199,7 @@ def shuffle(df, index, shuffle=None, npartitions=None, max_branch=32,
     df3 = rearrange_by_column(df2, '_partitions', npartitions=npartitions,
                               max_branch=max_branch, shuffle=shuffle,
                               compute=compute)
-
-    df4 = df3.drop('_partitions', axis=1, dtype=df.columns.dtype)
+    df4 = df3.map_partitions(drop_columns, '_partitions', df.columns.dtype)
     return df4
 
 
@@ -148,7 +211,9 @@ def rearrange_by_divisions(df, column, divisions, max_branch=None, shuffle=None)
     df2 = df.assign(_partitions=partitions)
     df3 = rearrange_by_column(df2, '_partitions', max_branch=max_branch,
                               npartitions=len(divisions) - 1, shuffle=shuffle)
-    return df3.drop('_partitions', axis=1, dtype=df.columns.dtype)
+    df4 = df3.drop('_partitions', axis=1)
+    df4 = df3.map_partitions(drop_columns, '_partitions', df.columns.dtype)
+    return df4
 
 
 def rearrange_by_column(df, col, npartitions=None, max_branch=None,
@@ -157,8 +222,6 @@ def rearrange_by_column(df, col, npartitions=None, max_branch=None,
     if shuffle == 'disk':
         return rearrange_by_column_disk(df, col, npartitions, compute=compute)
     elif shuffle == 'tasks':
-        if npartitions is not None and npartitions < df.npartitions:
-            raise ValueError("Must create as many or more partitions in shuffle")
         return rearrange_by_column_tasks(df, col, max_branch, npartitions)
     else:
         raise NotImplementedError("Unknown shuffle method %s" % shuffle)
@@ -167,18 +230,26 @@ def rearrange_by_column(df, col, npartitions=None, max_branch=None,
 class maybe_buffered_partd(object):
     """If serialized, will return non-buffered partd. Otherwise returns a
     buffered partd"""
-    def __init__(self, buffer=True):
+    def __init__(self, buffer=True, tempdir=None):
+        self.tempdir = tempdir or _globals.get('temporary_directory')
         self.buffer = buffer
 
     def __reduce__(self):
-        return (maybe_buffered_partd, (False,))
+        if self.tempdir:
+            return (maybe_buffered_partd, (False, self.tempdir))
+        else:
+            return (maybe_buffered_partd, (False,))
 
     def __call__(self, *args, **kwargs):
         import partd
-        if self.buffer:
-            return partd.PandasBlocks(partd.Buffer(partd.Dict(), partd.File()))
+        if self.tempdir:
+            file = partd.File(dir=self.tempdir)
         else:
-            return partd.PandasBlocks(partd.File())
+            file = partd.File()
+        if self.buffer:
+            return partd.PandasBlocks(partd.Buffer(partd.Dict(), file))
+        else:
+            return partd.PandasBlocks(file)
 
 
 def rearrange_by_column_disk(df, column, npartitions=None, compute=False):
@@ -280,8 +351,7 @@ def rearrange_by_column_tasks(df, column, max_branch=32, npartitions=None):
     df2 = DataFrame(dsk, 'shuffle-' + token, df, df.divisions)
 
     if npartitions is not None and npartitions != df.npartitions:
-        parts = partitioning_index(pd.Series(range(npartitions)),
-                                   df.npartitions)
+        parts = [i % df.npartitions for i in range(npartitions)]
         token = tokenize(df2, npartitions)
         dsk = {('repartition-group-' + token, i): (shuffle_group_2, k, column)
                for i, k in enumerate(df2._keys())}
@@ -304,7 +374,8 @@ def rearrange_by_column_tasks(df, column, max_branch=32, npartitions=None):
 
 
 def partitioning_index(df, npartitions):
-    """Computes a deterministic index mapping each record to a partition.
+    """
+    Computes a deterministic index mapping each record to a partition.
 
     Identical rows are mapped to the same partition.
 
@@ -319,35 +390,7 @@ def partitioning_index(df, npartitions):
     partitions : ndarray
         An array of int64 values mapping each record to a partition.
     """
-    if isinstance(df, (pd.Series, pd.Index)):
-        h = hash_series(df).astype('int64')
-    elif isinstance(df, pd.DataFrame):
-        cols = df.iteritems()
-        h = hash_series(next(cols)[1]).astype('int64')
-        for _, col in cols:
-            h = np.multiply(h, 3, h)
-            h = np.add(h, hash_series(col), h)
-    else:
-        raise TypeError("Unexpected type %s" % type(df))
-    return h % int(npartitions)
-
-
-def hash_series(s):
-    """Given a series, return a numpy array of deterministic integers."""
-    vals = s.values
-    dt = vals.dtype
-    if is_categorical_dtype(dt):
-        return vals.codes
-    elif np.issubdtype(dt, np.integer):
-        return vals
-    elif np.issubdtype(dt, np.floating):
-        return np.nan_to_num(vals).view('i' + str(dt.itemsize))
-    elif dt == np.bool:
-        return vals.view('int8')
-    elif np.issubdtype(dt, np.datetime64) or np.issubdtype(dt, np.timedelta64):
-        return vals.view('int64')
-    else:
-        return s.apply(hash).values
+    return hash_pandas_object(df, index=False) % int(npartitions)
 
 
 def barrier(args):
@@ -368,8 +411,16 @@ def set_partitions_pre(s, divisions):
 
 
 def shuffle_group_2(df, col):
-    g = df.groupby(col)
-    return {i: g.get_group(i) for i in g.groups}, df.head(0)
+    if not len(df):
+        return {}, df
+    ind = df[col]._values.astype(np.int64)
+    n = ind.max() + 1
+    indexer, locations = groupsort_indexer(ind.view(np.int64), n)
+    df2 = df.take(indexer)
+    locations = locations.cumsum()
+    parts = [df2.iloc[a:b] for a, b in zip(locations[:-1], locations[1:])]
+    result2 = dict(zip(range(n), parts))
+    return result2, df.iloc[:0]
 
 
 def shuffle_group_get(g_head, i):
@@ -381,10 +432,28 @@ def shuffle_group_get(g_head, i):
 
 
 def shuffle_group(df, col, stage, k, npartitions):
-    ind = partitioning_index(df[col], npartitions)
-    c = ind // k ** stage % k
-    g = df.groupby(c)
-    return {i: g.get_group(i) if i in g.groups else df.head(0) for i in range(k)}
+    if col == '_partitions':
+        ind = df[col]
+    else:
+        ind = hash_pandas_object(df[col], index=False)
+
+    c = ind._values
+    typ = np.min_scalar_type(npartitions * 2)
+    c = c.astype(typ)
+
+    npartitions, k, stage = [np.array(x, dtype=np.min_scalar_type(x))[()]
+                             for x in [npartitions, k, stage]]
+
+    c = np.mod(c, npartitions, out=c)
+    c = np.floor_divide(c, k ** stage, out=c)
+    c = np.mod(c, k, out=c)
+
+    indexer, locations = groupsort_indexer(c.astype(np.int64), k)
+    df2 = df.take(indexer)
+    locations = locations.cumsum()
+    parts = [df2.iloc[a:b] for a, b in zip(locations[:-1], locations[1:])]
+
+    return dict(zip(range(k), parts))
 
 
 def shuffle_group_3(df, col, npartitions, p):
@@ -404,3 +473,42 @@ def set_index_post_series(df, index_name, drop, column_dtype):
     df2.index.name = index_name
     df2.columns = df2.columns.astype(column_dtype)
     return df2
+
+
+def set_sorted_index(df, index, drop=True, divisions=None, **kwargs):
+    if not isinstance(index, Series):
+        meta = df._meta.set_index(index, drop=drop)
+    else:
+        meta = df._meta.set_index(index._meta, drop=drop)
+
+    result = map_partitions(M.set_index, df, index, drop=drop, meta=meta)
+
+    if not divisions:
+        divisions = compute_divisions(result, **kwargs)
+    elif len(divisions) != len(df.divisions):
+        msg = ("When doing `df.set_index(col, sorted=True, divisions=...)`, "
+               "divisions indicates known splits in the index column. In this "
+               "case divisions must be the same length as the existing "
+               "divisions in `df`\n\n"
+               "If the intent is to repartition into new divisions after "
+               "setting the index, you probably want:\n\n"
+               "`df.set_index(col, sorted=True).repartition(divisions=divisions)`")
+        raise ValueError(msg)
+
+    result.divisions = tuple(divisions)
+    return result
+
+
+def compute_divisions(df, **kwargs):
+    mins = df.index.map_partitions(M.min, meta=df.index)
+    maxes = df.index.map_partitions(M.max, meta=df.index)
+    mins, maxes = compute(mins, maxes, **kwargs)
+
+    if (sorted(mins) != list(mins) or
+            sorted(maxes) != list(maxes) or
+            any(a > b for a, b in zip(mins, maxes))):
+        raise ValueError("Partitions must be sorted ascending with the index",
+                         mins, maxes)
+
+    divisions = tuple(mins) + (list(maxes)[-1],)
+    return divisions
