@@ -32,11 +32,15 @@ def test_loc():
     assert_eq(d.loc[3:8], full.loc[3:8])
     assert_eq(d.loc[:8], full.loc[:8])
     assert_eq(d.loc[3:], full.loc[3:])
+    assert_eq(d.loc[[5]], full.loc[[5]])
+    assert_eq(d.loc[[3, 4, 1, 8]], full.loc[[3, 4, 1, 8]])
 
     assert_eq(d.a.loc[5], full.a.loc[5:5])
     assert_eq(d.a.loc[3:8], full.a.loc[3:8])
     assert_eq(d.a.loc[:8], full.a.loc[:8])
     assert_eq(d.a.loc[3:], full.a.loc[3:])
+    assert_eq(d.a.loc[[5]], full.a.loc[[5]])
+    assert_eq(d.a.loc[[3, 4, 1, 8]], full.a.loc[[3, 4, 1, 8]])
 
     pytest.raises(KeyError, lambda: d.loc[1000])
     assert_eq(d.loc[1000:], full.loc[1000:])
@@ -100,26 +104,45 @@ def test_loc2d():
     assert_eq(d.loc[3:8, 'a'], full.loc[3:8, 'a'])
     assert_eq(d.loc[:8, 'a'], full.loc[:8, 'a'])
     assert_eq(d.loc[3:, 'a'], full.loc[3:, 'a'])
+    assert_eq(d.loc[[8], 'a'], full.loc[[8], 'a'])
 
     assert_eq(d.loc[3:8, ['a']], full.loc[3:8, ['a']])
     assert_eq(d.loc[:8, ['a']], full.loc[:8, ['a']])
     assert_eq(d.loc[3:, ['a']], full.loc[3:, ['a']])
+    assert_eq(d.loc[[3, 4, 3], ['a']], full.loc[[3, 4, 3], ['a']])
 
     # 3d
-    with tm.assertRaises(pd.core.indexing.IndexingError):
+    with pytest.raises(pd.core.indexing.IndexingError):
         d.loc[3, 3, 3]
 
     # Series should raise
-    with tm.assertRaises(pd.core.indexing.IndexingError):
+    with pytest.raises(pd.core.indexing.IndexingError):
         d.a.loc[3, 3]
 
-    with tm.assertRaises(pd.core.indexing.IndexingError):
+    with pytest.raises(pd.core.indexing.IndexingError):
         d.a.loc[3:, 3]
 
-    with tm.assertRaises(pd.core.indexing.IndexingError):
+    with pytest.raises(pd.core.indexing.IndexingError):
         d.a.loc[d.a % 2 == 0, 3]
 
 
+def test_loc2d_with_known_divisions():
+    df = pd.DataFrame(np.random.randn(20, 5),
+                      index=list('abcdefghijklmnopqrst'),
+                      columns=list('ABCDE'))
+    ddf = dd.from_pandas(df, 3)
+
+    assert_eq(ddf.loc['a', 'A'], df.loc[['a'], 'A'])
+    assert_eq(ddf.loc['a', ['A']], df.loc[['a'], ['A']])
+    assert_eq(ddf.loc['a':'o', 'A'], df.loc['a':'o', 'A'])
+    assert_eq(ddf.loc['a':'o', ['A']], df.loc['a':'o', ['A']])
+    assert_eq(ddf.loc[['n'], ['A']], df.loc[['n'], ['A']])
+    assert_eq(ddf.loc[['a', 'c', 'n'], ['A']], df.loc[['a', 'c', 'n'], ['A']])
+    assert_eq(ddf.loc[['t', 'b'], ['A']], df.loc[['t', 'b'], ['A']])
+    assert_eq(ddf.loc[['r', 'r', 'c', 'g', 'h'], ['A']],
+              df.loc[['r', 'r', 'c', 'g', 'h'], ['A']])
+
+
 def test_loc2d_with_unknown_divisions():
     df = pd.DataFrame(np.random.randn(20, 5),
                       index=list('abcdefghijklmnopqrst'),
@@ -221,6 +244,20 @@ def test_loc_on_pandas_datetimes():
     assert_eq(a.loc['2014': '2015'], a.loc['2014': '2015'])
 
 
+def test_loc_datetime_no_freq():
+    # https://github.com/dask/dask/issues/2389
+
+    datetime_index = pd.date_range('2016-01-01', '2016-01-31', freq='12h')
+    datetime_index.freq = None  # FORGET FREQUENCY
+    df = pd.DataFrame({'num': range(len(datetime_index))}, index=datetime_index)
+
+    ddf = dd.from_pandas(df, npartitions=1)
+    slice_ = slice('2016-01-03', '2016-01-05')
+    result = ddf.loc[slice_, :]
+    expected = df.loc[slice_, :]
+    assert_eq(result, expected)
+
+
 def test_coerce_loc_index():
     for t in [pd.Timestamp, np.datetime64]:
         assert isinstance(_coerce_loc_index([t('2014')], '2014'), t)
