@@ -1,5 +1,6 @@
 from __future__ import absolute_import, division, print_function
 
+import re
 import textwrap
 from distutils.version import LooseVersion
 
@@ -11,10 +12,15 @@ from contextlib import contextmanager
 import numpy as np
 import pandas as pd
 import pandas.util.testing as tm
-from pandas.core.common import is_datetime64tz_dtype
-import toolz
+from pandas.api.types import is_categorical_dtype, is_scalar
+try:
+    from pandas.api.types import is_datetime64tz_dtype
+except ImportError:
+    # pandas < 0.19.2
+    from pandas.core.common import is_datetime64tz_dtype
 
-from ..async import get_sync
+from ..core import get_deps
+from ..local import get_sync
 
 
 PANDAS_VERSION = LooseVersion(pd.__version__)
@@ -58,7 +64,6 @@ def shard_df_on_index(df, divisions):
     3  30  2
     4  40  1
     """
-    from dask.dataframe.categorical import is_categorical_dtype
 
     if isinstance(divisions, Iterator):
         divisions = list(divisions)
@@ -77,30 +82,6 @@ def shard_df_on_index(df, divisions):
         yield df.iloc[indices[-1]:]
 
 
-def unique(divisions):
-    """ Polymorphic unique function
-
-    >>> list(unique([1, 2, 3, 1, 2, 3]))
-    [1, 2, 3]
-
-    >>> unique(np.array([1, 2, 3, 1, 2, 3]))
-    array([1, 2, 3])
-
-    >>> unique(pd.Categorical(['Alice', 'Bob', 'Alice'], ordered=False))
-    [Alice, Bob]
-    Categories (2, object): [Alice, Bob]
-    """
-    if isinstance(divisions, np.ndarray):
-        return np.unique(divisions)
-    if isinstance(divisions, pd.Categorical):
-        return pd.Categorical.from_codes(np.unique(divisions.codes),
-                                         divisions.categories,
-                                         divisions.ordered)
-    if isinstance(divisions, (tuple, list, Iterator)):
-        return tuple(toolz.unique(divisions))
-    raise NotImplementedError()
-
-
 _META_TYPES = "meta : pd.DataFrame, pd.Series, dict, iterable, tuple, optional"
 _META_DESCRIPTION = """\
 An empty ``pd.DataFrame`` or ``pd.Series`` that matches the dtypes and
@@ -123,12 +104,21 @@ def insert_meta_param_description(*args, **kwargs):
     if not args:
         return lambda f: insert_meta_param_description(f, **kwargs)
     f = args[0]
+    indent = " " * kwargs.get('pad', 8)
+    body = textwrap.wrap(_META_DESCRIPTION, initial_indent=indent,
+                         subsequent_indent=indent, width=78)
+    descr = '{0}\n{1}'.format(_META_TYPES, '\n'.join(body))
     if f.__doc__:
-        indent = " " * kwargs.get('pad', 8)
-        body = textwrap.wrap(_META_DESCRIPTION, initial_indent=indent,
-                             subsequent_indent=indent, width=78)
-        descr = '{0}\n{1}'.format(_META_TYPES, '\n'.join(body))
-        f.__doc__ = f.__doc__.replace('$META', descr)
+        if '$META' in f.__doc__:
+            f.__doc__ = f.__doc__.replace('$META', descr)
+        else:
+            # Put it at the end of the parameters section
+            parameter_header = 'Parameters\n%s----------' % indent[4:]
+            first, last = re.split('Parameters\\n[ ]*----------', f.__doc__)
+            parameters, rest = last.split('\n\n', 1)
+            f.__doc__ = '{0}{1}{2}\n{3}{4}\n\n{5}'.format(first, parameter_header,
+                                                          parameters, indent[4:],
+                                                          descr, rest)
     return f
 
 
@@ -159,6 +149,89 @@ def raise_on_meta_error(funcname=None):
         raise ValueError(msg)
 
 
+UNKNOWN_CATEGORIES = '__UNKNOWN_CATEGORIES__'
+
+
+def has_known_categories(x):
+    """Returns whether the categories in `x` are known.
+
+    Parameters
+    ----------
+    x : Series or CategoricalIndex
+    """
+    x = getattr(x, '_meta', x)
+    if isinstance(x, pd.Series):
+        return UNKNOWN_CATEGORIES not in x.cat.categories
+    elif isinstance(x, pd.CategoricalIndex):
+        return UNKNOWN_CATEGORIES not in x.categories
+    raise TypeError("Expected Series or CategoricalIndex")
+
+
+def strip_unknown_categories(x):
+    """Replace any unknown categoricals with empty categoricals.
+
+    Useful for preventing ``UNKNOWN_CATEGORIES`` from leaking into results.
+    """
+    if isinstance(x, (pd.Series, pd.DataFrame)):
+        x = x.copy()
+        if isinstance(x, pd.DataFrame):
+            cat_mask = x.dtypes == 'category'
+            if cat_mask.any():
+                cats = cat_mask[cat_mask].index
+                for c in cats:
+                    if not has_known_categories(x[c]):
+                        x[c].cat.set_categories([], inplace=True)
+        elif isinstance(x, pd.Series):
+            if is_categorical_dtype(x.dtype) and not has_known_categories(x):
+                x.cat.set_categories([], inplace=True)
+        if (isinstance(x.index, pd.CategoricalIndex) and not
+                has_known_categories(x.index)):
+            x.index = x.index.set_categories([])
+    elif isinstance(x, pd.CategoricalIndex) and not has_known_categories(x):
+        x = x.set_categories([])
+    return x
+
+
+def clear_known_categories(x, cols=None, index=True):
+    """Set categories to be unknown.
+
+    Parameters
+    ----------
+    x : DataFrame, Series, Index
+    cols : iterable, optional
+        If x is a DataFrame, set only categoricals in these columns to unknown.
+        By default, all categorical columns are set to unknown categoricals
+    index : bool, optional
+        If True and x is a Series or DataFrame, set the clear known categories
+        in the index as well.
+    """
+    if isinstance(x, (pd.Series, pd.DataFrame)):
+        x = x.copy()
+        if isinstance(x, pd.DataFrame):
+            mask = x.dtypes == 'category'
+            if cols is None:
+                cols = mask[mask].index
+            elif not mask.loc[cols].all():
+                raise ValueError("Not all columns are categoricals")
+            for c in cols:
+                x[c].cat.set_categories([UNKNOWN_CATEGORIES], inplace=True)
+        elif isinstance(x, pd.Series):
+            if is_categorical_dtype(x.dtype):
+                x.cat.set_categories([UNKNOWN_CATEGORIES], inplace=True)
+        if index and isinstance(x.index, pd.CategoricalIndex):
+            x.index = x.index.set_categories([UNKNOWN_CATEGORIES])
+    elif isinstance(x, pd.CategoricalIndex):
+        x = x.set_categories([UNKNOWN_CATEGORIES])
+    return x
+
+
+def _empty_series(name, dtype, index=None):
+    if isinstance(dtype, str) and dtype == 'category':
+        return pd.Series(pd.Categorical([UNKNOWN_CATEGORIES]),
+                         name=name, index=index).iloc[:0]
+    return pd.Series([], dtype=dtype, name=name, index=index)
+
+
 def make_meta(x, index=None):
     """Create an empty pandas object containing the desired metadata.
 
@@ -192,17 +265,17 @@ def make_meta(x, index=None):
     elif isinstance(x, pd.Index):
         return x[0:0]
     index = index if index is None else index[0:0]
+
     if isinstance(x, dict):
-        return pd.DataFrame({c: pd.Series([], dtype=d)
-                             for (c, d) in x.items()},
-                            index=index)
-    elif isinstance(x, tuple) and len(x) == 2:
-        return pd.Series([], dtype=x[1], name=x[0], index=index)
+        return pd.DataFrame({c: _empty_series(c, d, index=index)
+                             for (c, d) in x.items()}, index=index)
+    if isinstance(x, tuple) and len(x) == 2:
+        return _empty_series(x[0], x[1], index=index)
     elif isinstance(x, (list, tuple)):
         if not all(isinstance(i, tuple) and len(i) == 2 for i in x):
             raise ValueError("Expected iterable of tuples of (name, dtype), "
                              "got {0}".format(x))
-        return pd.DataFrame({c: pd.Series([], dtype=d) for (c, d) in x},
+        return pd.DataFrame({c: _empty_series(c, d, index=index) for (c, d) in x},
                             columns=[c for c, d in x], index=index)
     elif not hasattr(x, 'dtype') and x is not None:
         # could be a string, a dtype object, or a python type. Skip `None`,
@@ -215,23 +288,34 @@ def make_meta(x, index=None):
             # Continue on to next check
             pass
 
-    if is_pd_scalar(x):
+    if is_scalar(x):
         return _nonempty_scalar(x)
 
     raise TypeError("Don't know how to create metadata from {0}".format(x))
 
 
+if PANDAS_VERSION >= "0.20.0":
+    _numeric_index_types = (pd.Int64Index, pd.Float64Index, pd.UInt64Index)
+else:
+    _numeric_index_types = (pd.Int64Index, pd.Float64Index)
+
+
 def _nonempty_index(idx):
     typ = type(idx)
     if typ is pd.RangeIndex:
         return pd.RangeIndex(2, name=idx.name)
-    elif typ in (pd.Int64Index, pd.Float64Index):
+    elif typ in _numeric_index_types:
         return typ([1, 2], name=idx.name)
     elif typ is pd.Index:
         return pd.Index(['a', 'b'], name=idx.name)
     elif typ is pd.DatetimeIndex:
         start = '1970-01-01'
-        data = [start, start] if idx.freq is None else None
+        # Need a non-monotonic decreasing index to avoid issues with
+        # partial string indexing see https://github.com/dask/dask/issues/2389
+        # and https://github.com/pandas-dev/pandas/issues/16515
+        # This doesn't mean `_meta_nonempty` should ever rely on
+        # `self.monotonic_increasing` or `self.monotonic_decreasing`
+        data = [start, '1970-01-02'] if idx.freq is None else None
         return pd.DatetimeIndex(data, start=start, periods=2, freq=idx.freq,
                                 tz=idx.tz, name=idx.name)
     elif typ is pd.PeriodIndex:
@@ -243,9 +327,13 @@ def _nonempty_index(idx):
         return pd.TimedeltaIndex(data, start=start, periods=2, freq=idx.freq,
                                  name=idx.name)
     elif typ is pd.CategoricalIndex:
-        element = idx.categories[0]
-        return pd.CategoricalIndex([element, element],
-                                   categories=idx.categories,
+        if len(idx.categories):
+            data = [idx.categories[0]] * 2
+            cats = idx.categories
+        else:
+            data = _nonempty_index(idx.categories)
+            cats = None
+        return pd.CategoricalIndex(data, categories=cats,
                                    ordered=idx.ordered, name=idx.name)
     elif typ is pd.MultiIndex:
         levels = [_nonempty_index(i) for i in idx.levels]
@@ -290,25 +378,21 @@ def _nonempty_scalar(x):
                         "'{0}'".format(type(x).__name__))
 
 
-def is_pd_scalar(x):
-    """Whether the object is a scalar type"""
-    return (np.isscalar(x) or isinstance(x, (pd.Timestamp, pd.Timedelta,
-                                             pd.Period)))
-
-
 def _nonempty_series(s, idx):
 
-    from dask.dataframe.categorical import is_categorical_dtype
-
     dtype = s.dtype
     if is_datetime64tz_dtype(dtype):
         entry = pd.Timestamp('1970-01-01', tz=dtype.tz)
         data = [entry, entry]
     elif is_categorical_dtype(dtype):
-        entry = s.cat.categories[0]
-        data = pd.Categorical([entry, entry],
-                               categories=s.cat.categories,
-                               ordered=s.cat.ordered)
+        if len(s.cat.categories):
+            data = [s.cat.categories[0]] * 2
+            cats = s.cat.categories
+        else:
+            data = _nonempty_index(s.cat.categories)
+            cats = None
+        data = pd.Categorical(data, categories=cats,
+                              ordered=s.cat.ordered)
     else:
         entry = _scalar_from_dtype(dtype)
         data = np.array([entry, entry], dtype=dtype)
@@ -329,9 +413,13 @@ def meta_nonempty(x):
         return _nonempty_series(x, idx)
     elif isinstance(x, pd.DataFrame):
         idx = _nonempty_index(x.index)
-        data = {c: _nonempty_series(x[c], idx) for c in x.columns}
-        return pd.DataFrame(data, columns=x.columns, index=idx)
-    elif is_pd_scalar(x):
+        data = {i: _nonempty_series(x.iloc[:, i], idx)
+                for i, c in enumerate(x.columns)}
+        res = pd.DataFrame(data, index=idx,
+                           columns=np.arange(len(x.columns)))
+        res.columns = x.columns
+        return res
+    elif is_scalar(x):
         return _nonempty_scalar(x)
     else:
         raise TypeError("Expected Index, Series, DataFrame, or scalar, "
@@ -343,41 +431,42 @@ def meta_nonempty(x):
 ###############################################################
 
 
-def _check_dask(dsk, check_names=True, check_dtypes=True):
+def _check_dask(dsk, check_names=True, check_dtypes=True, result=None):
     import dask.dataframe as dd
     if hasattr(dsk, 'dask'):
-        result = dsk.compute(get=get_sync)
+        if result is None:
+            result = dsk.compute(get=get_sync)
         if isinstance(dsk, dd.Index):
             assert isinstance(result, pd.Index), type(result)
-            if check_names:
-                assert dsk.name == result.name
-            # cache
             assert isinstance(dsk._meta, pd.Index), type(dsk._meta)
             if check_names:
+                assert dsk.name == result.name
                 assert dsk._meta.name == result.name
+                if isinstance(result, pd.MultiIndex):
+                    assert result.names == dsk._meta.names
             if check_dtypes:
                 assert_dask_dtypes(dsk, result)
         elif isinstance(dsk, dd.Series):
             assert isinstance(result, pd.Series), type(result)
-            if check_names:
-                assert dsk.name == result.name, (dsk.name, result.name)
-            # cache
             assert isinstance(dsk._meta, pd.Series), type(dsk._meta)
             if check_names:
+                assert dsk.name == result.name, (dsk.name, result.name)
                 assert dsk._meta.name == result.name
             if check_dtypes:
                 assert_dask_dtypes(dsk, result)
+            _check_dask(dsk.index, check_names=check_names,
+                        check_dtypes=check_dtypes, result=result.index)
         elif isinstance(dsk, dd.DataFrame):
             assert isinstance(result, pd.DataFrame), type(result)
             assert isinstance(dsk.columns, pd.Index), type(dsk.columns)
-            if check_names:
-                tm.assert_index_equal(dsk.columns, result.columns)
-            # cache
             assert isinstance(dsk._meta, pd.DataFrame), type(dsk._meta)
             if check_names:
+                tm.assert_index_equal(dsk.columns, result.columns)
                 tm.assert_index_equal(dsk._meta.columns, result.columns)
             if check_dtypes:
                 assert_dask_dtypes(dsk, result)
+            _check_dask(dsk.index, check_names=check_names,
+                        check_dtypes=check_dtypes, result=result.index)
         elif isinstance(dsk, dd.core.Scalar):
             assert (np.isscalar(result) or
                     isinstance(result, (pd.Timestamp, pd.Timedelta)))
@@ -402,15 +491,22 @@ def _maybe_sort(a):
     return a.sort_index()
 
 
-def eq(a, b, check_names=True, check_dtypes=True, check_divisions=True,
-       **kwargs):
+def assert_eq(a, b, check_names=True, check_dtypes=True,
+              check_divisions=True, check_index=True, **kwargs):
     if check_divisions:
         assert_divisions(a)
         assert_divisions(b)
+        if hasattr(a, 'divisions') and hasattr(b, 'divisions'):
+            at = type(np.asarray(a.divisions).tolist()[0])  # numpy to python
+            bt = type(np.asarray(b.divisions).tolist()[0])  # scalar conversion
+            assert at == bt, (at, bt)
     assert_sane_keynames(a)
     assert_sane_keynames(b)
     a = _check_dask(a, check_names=check_names, check_dtypes=check_dtypes)
     b = _check_dask(b, check_names=check_names, check_dtypes=check_dtypes)
+    if not check_index:
+        a = a.reset_index(drop=True)
+        b = b.reset_index(drop=True)
     if isinstance(a, pd.DataFrame):
         a = _maybe_sort(a)
         b = _maybe_sort(b)
@@ -510,3 +606,11 @@ def assert_dask_dtypes(ddf, res, numeric_equal=True):
             assert (a.kind in eq_types and b.kind in eq_types) or (a == b)
         else:
             assert type(ddf._meta) == type(res)
+
+
+def assert_max_deps(x, n, eq=True):
+    dependencies, dependents = get_deps(x.dask)
+    if eq:
+        assert max(map(len, dependencies.values())) == n
+    else:
+        assert max(map(len, dependencies.values())) <= n
