@@ -1,5 +1,7 @@
 from __future__ import absolute_import, division, print_function
 
+import random
+from bisect import bisect_left
 from itertools import cycle
 from operator import itemgetter, add
 
@@ -119,7 +121,8 @@ def get_colors(palette, funcs):
     Parameters
     ----------
     palette : string
-        Name of the palette. Must be an attribute on bokeh.palettes.
+        Name of the bokeh palette to use, must be a member of
+        bokeh.palettes.all_palettes.
     funcs : iterable
         Iterable of function names
     """
@@ -128,16 +131,16 @@ def get_colors(palette, funcs):
 
     unique_funcs = list(sorted(tz.unique(funcs)))
     n_funcs = len(unique_funcs)
-    palette_lookup = getattr(palettes, palette)
-    keys = list(palette_lookup.keys())
-    low, high = min(keys), max(keys)
-    if n_funcs > high:
-        colors = cycle(palette_lookup[high])
-    elif n_funcs < low:
-        colors = palette_lookup[low]
-    else:
-        colors = palette_lookup[n_funcs]
-    color_lookup = dict(zip(unique_funcs, colors))
+    palette_lookup = palettes.all_palettes[palette]
+    keys = list(sorted(palette_lookup.keys()))
+    index = keys[min(bisect_left(keys, n_funcs), len(keys) - 1)]
+    palette = palette_lookup[index]
+    # Some bokeh palettes repeat colors, we want just the unique set
+    palette = list(tz.unique(palette))
+    if len(palette) > n_funcs:
+        # Consistently shuffle palette - prevents just using low-range
+        random.Random(42).shuffle(palette)
+    color_lookup = dict(zip(unique_funcs, cycle(palette)))
     return [color_lookup[n] for n in funcs]
 
 
@@ -216,8 +219,8 @@ def plot_tasks(results, dsk, palette='Viridis', label_size=60, **kwargs):
     dsk : dict
         The dask graph being profiled.
     palette : string, optional
-        Name of the bokeh palette to use, must be an attribute on
-        bokeh.palettes.
+        Name of the bokeh palette to use, must be a member of
+        bokeh.palettes.all_palettes.
     label_size: int (optional)
         Maximum size of output labels in plot, defaults to 60
     **kwargs
@@ -298,8 +301,8 @@ def plot_resources(results, palette='Viridis', **kwargs):
     results : sequence
         Output of ResourceProfiler.results
     palette : string, optional
-        Name of the bokeh palette to use, must be an attribute on
-        bokeh.palettes.
+        Name of the bokeh palette to use, must be a member of
+        bokeh.palettes.all_palettes.
     **kwargs
         Other keyword arguments, passed to bokeh.figure. These will override
         all defaults set by plot_resources.
@@ -325,7 +328,7 @@ def plot_resources(results, palette='Viridis', **kwargs):
     else:
         t = mem = cpu = []
         p = bp.figure(y_range=(0, 100), x_range=(0, 10), **defaults)
-    colors = getattr(palettes, palette)[6]
+    colors = palettes.all_palettes[palette][6]
     p.line(t, cpu, color=colors[0], line_width=4, legend='% CPU')
     p.yaxis.axis_label = "% CPU"
     p.extra_y_ranges = {'memory': Range1d(start=(min(mem) if mem else 0),
@@ -353,8 +356,8 @@ def plot_cache(results, dsk, start_time, metric_name, palette='Viridis',
     metric_name : string
         Metric used to measure cache size
     palette : string, optional
-        Name of the bokeh palette to use, must be an attribute on
-        bokeh.palettes.
+        Name of the bokeh palette to use, must be a member of
+        bokeh.palettes.all_palettes.
     label_size: int (optional)
         Maximum size of output labels in plot, defaults to 60
     **kwargs
