@@ -3,14 +3,17 @@
 import os
 import pytest
 from operator import add, mul
+import subprocess
 import sys
 
 import dask
+from dask import delayed
 from dask.base import (compute, tokenize, normalize_token, normalize_function,
-                       visualize)
+                       visualize, persist, function_cache)
+from dask.delayed import Delayed
 from dask.utils import tmpdir, tmpfile, ignoring
 from dask.utils_test import inc, dec
-from dask.compatibility import unicode
+from dask.compatibility import long, unicode
 
 
 def import_or_none(path):
@@ -18,6 +21,7 @@ def import_or_none(path):
         return pytest.importorskip(path)
     return None
 
+
 tz = pytest.importorskip('toolz')
 da = import_or_none('dask.array')
 db = import_or_none('dask.bag')
@@ -151,8 +155,19 @@ def test_tokenize_numpy_memmap_no_filename():
         assert tokenize(b) == tokenize(b)
 
 
+@pytest.mark.skipif('not np')
+def test_tokenize_numpy_ufunc_consistent():
+    assert tokenize(np.sin) == '02106e2c67daf452fb480d264e0dac21'
+    assert tokenize(np.cos) == 'c99e52e912e4379882a9a4b387957a0b'
+
+    # Make a ufunc that isn't in the numpy namespace. Similar to
+    # any found in other packages.
+    inc = np.frompyfunc(lambda x: x + 1, 1, 1)
+    assert tokenize(inc) == tokenize(inc)
+
+
 def test_normalize_base():
-    for i in [1, 1.1, '1', slice(1, 2, 3)]:
+    for i in [1, long(1), 1.1, '1', slice(1, 2, 3)]:
         assert normalize_token(i) is i
 
 
@@ -208,6 +223,10 @@ def test_tokenize_dict():
     assert tokenize({'x': 1, 1: 'x'}) == tokenize({'x': 1, 1: 'x'})
 
 
+def test_tokenize_set():
+    assert tokenize({1, 2, 'x', (1, 'x')}) == tokenize({1, 2, 'x', (1, 'x')})
+
+
 def test_tokenize_ordered_dict():
     with ignoring(ImportError):
         from collections import OrderedDict
@@ -219,6 +238,18 @@ def test_tokenize_ordered_dict():
         assert tokenize(a) != tokenize(c)
 
 
+@pytest.mark.skipif('not np')
+def test_tokenize_object_array_with_nans():
+    a = np.array([u'foo', u'Jos\xe9', np.nan], dtype='O')
+    assert tokenize(a) == tokenize(a)
+
+
+@pytest.mark.parametrize('x', [1, True, 'a', b'a', 1.0, 1j, 1.0j,
+                               [], (), {}, None, str, int])
+def test_tokenize_base_types(x):
+    assert tokenize(x) == tokenize(x), x
+
+
 @pytest.mark.skipif('not db')
 def test_compute_no_opt():
     # Bag does `fuse` by default. Test that with `optimize_graph=False` that
@@ -239,8 +270,11 @@ def test_compute_no_opt():
     keys = []
     with Callback(pretask=lambda key, *args: keys.append(key)):
         o.compute(get=dask.get)
-    assert len([k for k in keys if 'mul' in k[0]]) == 4
-    assert len([k for k in keys if 'add' in k[0]]) == 0
+    # Names of fused tasks have been merged, and the original key is an alias.
+    # Otherwise, the lengths below would be 4 and 0.
+    assert len([k for k in keys if 'mul' in k[0]]) == 8
+    assert len([k for k in keys if 'add' in k[0]]) == 4
+    assert len([k for k in keys if 'add-map-mul' in k[0]]) == 4  # See? Renamed
 
 
 @pytest.mark.skipif('not da')
@@ -254,6 +288,19 @@ def test_compute_array():
     assert np.allclose(out2, arr + 2)
 
 
+@pytest.mark.skipif('not da')
+def test_persist_array():
+    from dask.array.utils import assert_eq
+    arr = np.arange(100).reshape((10, 10))
+    x = da.from_array(arr, chunks=(5, 5))
+    x = (x + 1) - x.mean(axis=0)
+    y = x.persist()
+
+    assert_eq(x, y)
+    assert set(y.dask).issubset(x.dask)
+    assert len(y.dask) == y.npartitions
+
+
 @pytest.mark.skipif('not dd')
 def test_compute_dataframe():
     df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [5, 5, 3, 3]})
@@ -283,7 +330,7 @@ def test_compute_array_bag():
 
     pytest.raises(ValueError, lambda: compute(x, b))
 
-    xx, bb = compute(x, b, get=dask.async.get_sync)
+    xx, bb = compute(x, b, get=dask.get)
     assert np.allclose(xx, np.arange(5))
     assert bb == [1, 2, 3]
 
@@ -300,6 +347,19 @@ def test_compute_with_literal():
     assert compute(5) == (5,)
 
 
+def test_compute_nested():
+    a = delayed(1) + 5
+    b = a + 1
+    c = a + 2
+    assert (compute({'a': a, 'b': [1, 2, b]}, (c, 2)) ==
+            ({'a': 6, 'b': [1, 2, 7]}, (8, 2)))
+
+    res = compute([a, b], c, traverse=False)
+    assert res[0][0] is a
+    assert res[0][1] is b
+    assert res[1] == 8
+
+
 @pytest.mark.skipif('not da')
 @pytest.mark.skipif(sys.flags.optimize == 2,
                     reason="graphviz exception with Python -OO flag")
@@ -349,3 +409,108 @@ def test_optimizations_keyword():
         assert x.compute() == 0
 
     assert x.compute() == 2
+
+
+def test_default_imports():
+    """
+    Startup time: `import dask` should not import too many modules.
+    """
+    code = """if 1:
+        import dask
+        import sys
+
+        print(sorted(sys.modules))
+        """
+
+    out = subprocess.check_output([sys.executable, '-c', code])
+    modules = set(eval(out.decode()))
+    assert 'dask' in modules
+    blacklist = ['dask.array', 'dask.dataframe', 'numpy', 'pandas',
+                 'partd', 's3fs', 'distributed']
+    for mod in blacklist:
+        assert mod not in modules
+
+
+def test_persist_literals():
+    assert persist(1, 2, 3) == (1, 2, 3)
+
+
+def test_persist_delayed():
+    x1 = delayed(1)
+    x2 = delayed(inc)(x1)
+    x3 = delayed(inc)(x2)
+
+    xx, = persist(x3)
+    assert isinstance(xx, Delayed)
+    assert xx.key == x3.key
+    assert len(xx.dask) == 1
+
+    assert x3.compute() == xx.compute()
+
+
+@pytest.mark.skipif('not da or not db')
+def test_persist_array_bag():
+    x = da.arange(5, chunks=2) + 1
+    b = db.from_sequence([1, 2, 3]).map(inc)
+
+    with pytest.raises(ValueError):
+        persist(x, b)
+
+    xx, bb = persist(x, b, get=dask.get)
+
+    assert isinstance(xx, da.Array)
+    assert isinstance(bb, db.Bag)
+
+    assert xx.name == x.name
+    assert bb.name == b.name
+    assert len(xx.dask) == xx.npartitions < len(x.dask)
+    assert len(bb.dask) == bb.npartitions < len(b.dask)
+
+    assert np.allclose(x, xx)
+    assert list(b) == list(bb)
+
+
+def test_normalize_function_limited_size():
+    for i in range(1000):
+        normalize_function(lambda x: x)
+
+    assert 50 < len(function_cache) < 600
+
+
+def test_optimize_globals():
+    da = pytest.importorskip('dask.array')
+    db = pytest.importorskip('dask.bag')
+
+    x = da.ones(10, chunks=(5,))
+
+    def optimize_double(dsk, keys):
+        return {k: (mul, 2, v) for k, v in dsk.items()}
+
+    from dask.array.utils import assert_eq
+
+    assert_eq(x + 1, np.ones(10) + 1)
+
+    with dask.set_options(array_optimize=optimize_double):
+        assert_eq(x + 1, (np.ones(10) * 2 + 1) * 2)
+
+    assert_eq(x + 1, np.ones(10) + 1)
+
+    b = db.range(10, npartitions=2)
+
+    with dask.set_options(array_optimize=optimize_double):
+        xx, bb = dask.compute(x + 1, b.map(inc), get=dask.get)
+        assert_eq(xx, (np.ones(10) * 2 + 1) * 2)
+
+
+def test_optimize_None():
+    da = pytest.importorskip('dask.array')
+
+    x = da.ones(10, chunks=(5,))
+    y = x[:9][1:8][::2] + 1  # normally these slices would be fused
+
+    def my_get(dsk, keys):
+        assert dsk == dict(y.dask)  # but they aren't
+        return dask.get(dsk, keys)
+
+    with dask.set_options(array_optimize=None, get=my_get):
+        y.compute()
