@@ -1,7 +1,8 @@
-from dask.context import set_options, _globals
+from dask.context import set_options, _globals, defer_to_globals
 import dask.array as da
 import dask
 
+
 def test_with_get():
     var = [0]
 
@@ -33,3 +34,23 @@ def test_set_options_context_manger():
         assert _globals['foo'] == 'baz'
     finally:
         del _globals['foo']
+
+
+def test_defer_to_globals():
+    @defer_to_globals('f')
+    def f():
+        return 1
+
+    assert f() == 1
+
+    with dask.set_options(f=lambda: 2):
+        assert f() == 2
+    assert f() == 1
+
+    @defer_to_globals('f', falsey=lambda: 3)
+    def f():
+        return 1
+
+    with dask.set_options(f=None):
+        assert f() == 3
+    assert f() == 1
